import inspect
import sys
from datetime import datetime
from decimal import Decimal, DecimalTuple
from functools import lru_cache
from operator import attrgetter
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from mode.utils.objects import cached_property
from mode.utils.text import pluralize

from faust.exceptions import ValidationError
from faust.types.models import (
    FieldDescriptorT,
    ModelT,
    T,
)
from faust.utils import iso8601

from .tags import Tag
from .typing import NodeType, TypeExpression

__all__ = [
    'TYPE_TO_FIELD',
    'FieldDescriptor',
    'BooleanField',
    'NumberField',
    'FloatField',
    'IntegerField',
    'DatetimeField',
    'DecimalField',
    'BytesField',
    'StringField',
    'field_for_type',
]

CharacterType = TypeVar('CharacterType', str, bytes)


def _is_concrete_model(typ: Type = None) -> bool:
    return (typ is not None and
            inspect.isclass(typ) and
            issubclass(typ, ModelT) and
            typ is not ModelT and
            not getattr(typ, '__is_abstract__', False))


class FieldDescriptor(FieldDescriptorT[T]):
    """Describes a field.

    Used for every field in Record so that they can be used in join's
    /group_by etc.

    Examples:
        >>> class Withdrawal(Record):
        ...    account_id: str
        ...    amount: float = 0.0

        >>> Withdrawal.account_id
        <FieldDescriptor: Withdrawal.account_id: str>
        >>> Withdrawal.amount
        <FieldDescriptor: Withdrawal.amount: float = 0.0>

    Arguments:
        field (str): Name of field.
        type (Type): Field value type.
        required (bool): Set to false if field is optional.
        default (Any): Default value when `required=False`.

    Keyword Arguments:
        model (Type): Model class the field belongs to.
        parent (FieldDescriptorT): parent field if any.
    """

    #: Name of attribute on Model.
    field: str

    #: Name of field in serialized data (if differs from :attr:`field`).
    #: Defaults to :attr:`field`.
    input_name: str

    #: Name of field when serializing data (if differs from :attr:`field`).
    #: Defaults to :attr:`input_name`.
    output_name: str

    #: Type of value (e.g. ``int``, or ``Optional[int]``)).
    type: Type[T]

    #: The model class this field is associated with.
    model: Type[ModelT]

    #: Set if a value for this field is required (cannot be :const:`None`).
    required: bool = True

    #: Default value for non-required field.
    default: Optional[T] = None  # noqa: E704

    #: Coerce value to field descriptors type.
    #: This means assigning a value to this field, will first convert
    #: the value to the requested type.
    #: For example for a :class:`FloatField` the input will be converted
    #: to float, and passing any value that cannot be converted to float
    #: will raise an error.
    #:
    #: If coerce is not enabled you can store any type of value.
    #:
    #: Note: :const:`None` is usually considered a valid value for any field
    #: but this depends on the descriptor type.
    coerce: bool = False

    #: Exclude field from model representation.
    #: This means the field will not be part of the serialized structure.
    #: (``Model.dumps()``, ``Model.asdict()``, and
    #: ``Model.to_representation()``).
    exclude: bool = False

    #: Field may be tagged with Secret/Sensitive/etc.
    tag: Optional[Type[Tag]]

    _to_python: Optional[Callable[[T], T]]
    _expr: Optional[TypeExpression]

    def __init__(self, *,
                 field: str = None,
                 input_name: str = None,
                 output_name: str = None,
                 type: Type[T] = None,
                 model: Type[ModelT] = None,
                 required: bool = True,
                 default: T = None,
                 parent: FieldDescriptorT = None,
                 coerce: bool = None,
                 exclude: bool = None,
                 date_parser: Callable[[Any], datetime] = None,
                 tag: Type[Tag] = None,
                 **options: Any) -> None:
        self.field = cast(str, field)
        self.input_name = cast(str, input_name or field)
        self.output_name = output_name or self.input_name
        self.type = cast(Type[T], type)
        self.model = cast(Type[ModelT], model)
        self.required = required
        self.default = default
        self.parent = parent
        self._copy_descriptors(self.type)
        if coerce is not None:
            self.coerce = coerce
        if exclude is not None:
            self.exclude = exclude
        self.options = options
        if date_parser is None:
            date_parser = iso8601.parse
        self.date_parser: Callable[[Any], datetime] = date_parser
        self.tag = tag
        self._to_python = None
        self._expr = None

    def on_model_attached(self) -> None:
        self._expr = self._prepare_type_expression()
        self._to_python = self._compile_type_expression()

    def _prepare_type_expression(self) -> TypeExpression:
        assert self.model is not None
        expr = TypeExpression(
            self.type,
            user_types=self.model._options.coercions,
            date_parser=self.date_parser,
        )
        return expr

    def _compile_type_expression(self) -> Optional[Callable[[T], T]]:
        assert self._expr is not None
        expr = self._expr
        comprehension = expr.as_function(stacklevel=2)
        if (expr.has_generic_types and
                expr.has_custom_types) or expr.has_nonfield_types:
            return cast(Callable, comprehension)
        return None

    def __set_name__(self, owner: Type[ModelT], name: str) -> None:
        self.model = owner
        self.field = name

    def clone(self, **kwargs: Any) -> FieldDescriptorT:
        return type(self)(**{**self.as_dict(), **kwargs})

    def as_dict(self) -> Mapping[str, Any]:
        return {
            'field': self.field,
            'input_name': self.input_name,
            'output_name': self.output_name,
            'type': self.type,
            'model': self.model,
            'required': self.required,
            'default': self.default,
            'parent': self.parent,
            'coerce': self.coerce,
            'exclude': self.exclude,
            'date_parser': self.date_parser,
            'tag': self.tag,
            **self.options,
        }

    def validate_all(self, value: Any) -> Iterable[ValidationError]:
        need_coercion = not self.coerce
        try:
            v = self.prepare_value(value, coerce=need_coercion)
        except (TypeError, ValueError) as exc:
            vt = type(value)
            yield self.validation_error(
                f'{self.field} is not correct type for {self}, '
                f'got {vt!r}: {exc!r}')
        except Exception as exc:
            yield self.validation_error(
                f'{self.field} got internal error for value {value!r} '
                f'{exc!r}')
        else:
            if v is not None or self.required:
                yield from self.validate(cast(T, v))

    def validate(self, value: T) -> Iterable[ValidationError]:
        return iter([])

    def to_python(self, value: Any) -> Optional[T]:
        to_python = self._to_python
        if to_python is not None:
            value = to_python(value)
        return self.prepare_value(value)

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[T]:
        return cast(T, value)

    def _copy_descriptors(self, typ: Type = None) -> None:
        if typ is not None and _is_concrete_model(typ):
            typ._contribute_field_descriptors(self, typ._options, parent=self)

    def __get__(self, instance: Any, owner: Type) -> Any:
        # class attribute accessed
        if instance is None:
            return self

        field = self.field
        instance_dict = instance.__dict__
        to_python = self._to_python
        value = instance_dict[field]
        if self.lazy_coercion and to_python is not None:
            evaluated_fields: Set[str]
            evaluated_fields = instance.__evaluated_fields__
            if field not in evaluated_fields:
                if value is not None or self.required:
                    value = instance_dict[field] = to_python(value)
                evaluated_fields.add(field)
        return value

    def should_coerce(self, value: Any, coerce: bool = None) -> bool:
        c = coerce if coerce is not None else self.coerce
        return c and (self.required or value is not None)

    def getattr(self, obj: ModelT) -> T:
        """Get attribute from model recursively.

        Supports recursive lookups e.g. ``model.getattr('x.y.z')``.
        """
        return attrgetter('.'.join(reversed(list(self._parents_path()))))(obj)

    def _parents_path(self) -> Iterable[str]:
        node: Optional[FieldDescriptorT] = self
        while node:
            yield node.field
            node = node.parent

    def validation_error(self, reason: str) -> ValidationError:
        return ValidationError(reason, field=self)

    def __set__(self, instance: Any, value: T) -> None:
        value = cast(T, self.prepare_value(value))
        if self.tag:
            value = cast(T, self.tag(value, field=self.field))
        else:
            value = value
        instance.__dict__[self.field] = value

    def __repr__(self) -> str:
        default = '' if self.required else f' = {self.default!r}'
        typ = getattr(self.type, '__name__', self.type)
        return f'<{type(self).__name__}: {self.ident}: {typ}{default}>'

    @property
    def ident(self) -> str:
        """Return the fields identifier."""
        return f'{self.model.__name__}.{self.field}'

    @cached_property
    def related_models(self) -> Set[Type[ModelT]]:
        assert self._expr is not None
        return self._expr.found_types[NodeType.MODEL]

    @cached_property
    def lazy_coercion(self) -> bool:
        assert self._expr is not None
        return self._expr.has_generic_types or self._expr.has_models


class BooleanField(FieldDescriptor[bool]):

    def validate(self, value: T) -> Iterable[ValidationError]:
        if not isinstance(value, bool):
            yield self.validation_error(
                f'{self.field} must be True or False, of type bool')

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[bool]:
        if self.should_coerce(value, coerce):
            return True if value else False
        return value


class NumberField(FieldDescriptor[T]):

    max_value: Optional[int]
    min_value: Optional[int]

    def __init__(self, *,
                 max_value: int = None,
                 min_value: int = None,
                 **kwargs: Any) -> None:
        self.max_value = max_value
        self.min_value = min_value

        super().__init__(**kwargs, **{
            'max_value': max_value,
            'min_value': min_value,
        })

    def validate(self, value: T) -> Iterable[ValidationError]:
        val = cast(int, value)
        max_ = self.max_value
        if max_:
            if val > max_:
                yield self.validation_error(
                    f'{self.field} cannot be more than {max_}')
        min_ = self.min_value
        if min_:
            if val < min_:
                yield self.validation_error(
                    f'{self.field} must be at least {min_}')


class IntegerField(NumberField[int]):

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[int]:
        return int(value) if self.should_coerce(value, coerce) else value


class FloatField(NumberField[float]):

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[float]:
        return float(value) if self.should_coerce(value, coerce) else value


class DecimalField(NumberField[Decimal]):
    max_digits: Optional[int] = None
    max_decimal_places: Optional[int] = None

    def __init__(self, *,
                 max_digits: int = None,
                 max_decimal_places: int = None,
                 **kwargs: Any) -> None:
        self.max_digits = max_digits
        self.max_decimal_places = max_decimal_places

        super().__init__(**kwargs, **{
            'max_digits': max_digits,
            'max_decimal_places': max_decimal_places,
        })

    def to_python(self, value: Any) -> Any:
        if self._to_python is None:
            if self.model._options.decimals:
                return self.prepare_value(value, coerce=True)
            return self.prepare_value(value)
        else:
            return self._to_python(value)

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[Decimal]:
        return Decimal(value) if self.should_coerce(value, coerce) else value

    def validate(self, value: Decimal) -> Iterable[ValidationError]:
        if not value.is_finite():  # check for Inf/NaN/sNaN/qNaN
            yield self.validation_error(f'Illegal value in decimal: {value!r}')

        decimal_tuple: Optional[DecimalTuple] = None

        mdp = self.max_decimal_places
        if mdp:
            decimal_tuple = value.as_tuple()
            if abs(decimal_tuple.exponent) > mdp:
                yield self.validation_error(
                    f'{self.field} must have less than {mdp} decimal places.')
        max_digits = self.max_digits
        if max_digits:
            if decimal_tuple is None:
                decimal_tuple = value.as_tuple()
            digits = len(decimal_tuple.digits[:decimal_tuple.exponent])
            if digits > max_digits:
                yield self.validation_error(
                    f'{self.field} must have less than {max_digits} digits.')


class CharField(FieldDescriptor[CharacterType]):

    max_length: Optional[int]
    min_length: Optional[int]
    trim_whitespace: bool
    allow_blank: bool

    def __init__(self, *,
                 max_length: int = None,
                 min_length: int = None,
                 trim_whitespace: bool = False,
                 allow_blank: bool = False,
                 **kwargs: Any) -> None:
        self.max_length = max_length
        self.min_length = min_length
        self.trim_whitespace = trim_whitespace
        self.allow_blank = allow_blank
        super().__init__(
            **kwargs,
            **{'max_length': max_length,
               'min_length': min_length,
               'trim_whitespace': trim_whitespace,
               'allow_blank': allow_blank},
        )

    def validate(self, value: CharacterType) -> Iterable[ValidationError]:
        allow_blank = self.allow_blank
        if not allow_blank and not len(value):
            yield self.validation_error(f'{self.field} cannot be left blank')
        max_ = self.max_length
        length = len(value)
        min_ = self.min_length
        if min_:
            if length < min_:
                chars = pluralize(min_, 'character')
                yield self.validation_error(
                    f'{self.field} must have at least {min_} {chars}')
        if max_:
            if length > max_:
                chars = pluralize(max_, 'character')
                yield self.validation_error(
                    f'{self.field} must be at least {max_} {chars}')


class StringField(CharField[str]):

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[str]:
        if self.should_coerce(value, coerce):
            val = str(value) if not isinstance(value, str) else value
            if self.trim_whitespace:
                return val.strip()
            return val
        else:
            return value


class DatetimeField(FieldDescriptor[datetime]):

    def to_python(self, value: Any) -> Any:
        if self._to_python is None:
            if self.model._options.isodates:
                return self.prepare_value(value, coerce=True)
            return self.prepare_value(value)
        else:
            return self._to_python(value)

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[datetime]:
        if self.should_coerce(value, coerce):
            if value is not None and not isinstance(value, datetime):
                return self.date_parser(value)
            else:
                return value
        else:
            return value


class BytesField(CharField[bytes]):
    encoding: str = sys.getdefaultencoding()
    errors: str = 'strict'

    def __init__(self, *,
                 encoding: str = None,
                 errors: str = None,
                 **kwargs: Any) -> None:
        if encoding is not None:
            self.encoding = encoding
        if errors is not None:
            self.errors = errors
        super().__init__(
            encoding=self.encoding,
            errors=self.errors,
            **kwargs,
        )

    def prepare_value(self, value: Any, *,
                      coerce: bool = None) -> Optional[bytes]:
        if self.should_coerce(value, coerce):
            if isinstance(value, bytes):
                val = value
            else:
                val = cast(str, value).encode(encoding=self.encoding)
            if self.trim_whitespace:
                return val.strip()
            return val
        else:
            return value


TYPE_TO_FIELD = {
    bool: BooleanField,
    int: IntegerField,
    float: FloatField,
    Decimal: DecimalField,
    str: StringField,
    bytes: BytesField,
    datetime: DatetimeField,
}


@lru_cache(maxsize=2048)
def field_for_type(
        typ: Type) -> Tuple[Type[FieldDescriptorT], Optional[Type[Tag]]]:
    try:
        # 1) Check if type is in fast index.
        return TYPE_TO_FIELD[typ], None
    except KeyError:
        # 2) Unwrap Tag fields to their destination type.
        try:
            origin = typ.__origin__
        except AttributeError:
            pass
        else:
            try:
                if origin is not None and issubclass(origin, Tag):
                    return field_for_type(typ.__args__[0])[0], typ
            except TypeError:
                pass

        # 3) Check if type is a subclass of any of the types
        # in TYPE_TO_FIELD
        for basecls, DescriptorType in TYPE_TO_FIELD.items():
            try:
                if issubclass(typ, basecls):
                    return DescriptorType, None
            except TypeError:
                # not a type that can be used with issubclass
                # so skip trying
                break
        return FieldDescriptor, None



import asyncio
import sys
from pathlib import Path

import aiohttp_cors
import pytest
from aiohttp.web import Application
from mode.utils.mocks import ANY, AsyncMock, Mock, call, patch
from yarl import URL

from faust.types.web import ResourceOptions
from faust.web import base
from faust.web.drivers.aiohttp import (
    NON_OPTIONS_METHODS,
    Server,
    ServerThread,
    _prepare_cors_options,
)

if sys.platform == 'win32':
    DATAPATH = 'c:/Program Files/Faust/data'
else:
    DATAPATH = '/opt/data'


@pytest.fixture
def thread(*, web):
    return ServerThread(web)


@pytest.fixture
def server(*, web):
    return Server(web)


def test__prepare_cors_options():
    x = ResourceOptions(
        allow_credentials=True,
        expose_headers=['foo', 'bar'],
        allow_headers='*',
        max_age=13124,
        allow_methods='*',
    )
    y = ResourceOptions(
        allow_credentials=False,
        expose_headers='*',
        allow_headers='*',
        max_age=None,
        allow_methods=['POST'],
    )
    z = aiohttp_cors.ResourceOptions(allow_credentials=False)
    opts = {
        'http://bob.example.com': x,
        'http://alice.example.com': y,
        'http://moo.example.com': z,
    }

    aiohttp_cors_opts = _prepare_cors_options(opts)
    x1 = aiohttp_cors_opts['http://bob.example.com']
    assert isinstance(x1, aiohttp_cors.ResourceOptions)
    x2 = aiohttp_cors_opts['http://alice.example.com']
    assert isinstance(x2, aiohttp_cors.ResourceOptions)
    assert aiohttp_cors_opts['http://moo.example.com']

    assert x1.allow_credentials
    assert x1.expose_headers == {'foo', 'bar'}
    assert x1.allow_headers == '*'
    assert x1.max_age == 13124
    assert x1.allow_methods == '*'

    assert not x2.allow_credentials
    assert x2.expose_headers == '*'
    assert x2.allow_headers == '*'
    assert x2.max_age is None
    assert x2.allow_methods == {'POST'}


class test_ServerThread:

    def test_constructor(self, *, thread, web):
        assert thread.web is web

    @pytest.mark.asyncio
    async def test_on_start(self, *, thread):
        thread.web.start_server = AsyncMock(name='web.start_server')
        await thread.on_start()

        thread.web.start_server.assert_called_once()

    @pytest.mark.asyncio
    async def test_crash(self, *, thread):
        thread._thread_running = asyncio.Future()
        exc = RuntimeError()
        await thread.crash(exc)
        assert thread._thread_running.exception() is exc

        thread._thread_running = None
        await thread.crash(exc)

    @pytest.mark.asyncio
    async def test_on_thread_start(self, *, thread):
        thread.web = Mock(
            name='web',
            autospec=base.Web,
            stop_server=AsyncMock(),
        )
        await thread.on_thread_stop()

        thread.web.stop_server.assert_called_once()


class test_Server:

    def test_constructor(self, *, server, web):
        assert server.web is web

    @pytest.mark.asyncio
    async def test_on_start(self, *, server, web):
        web.start_server = AsyncMock()
        await server.on_start()
        web.start_server.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_stop(self, *, server, web):
        web.stop_server = AsyncMock()
        await server.on_stop()
        web.stop_server.assert_called_once_with()


class test_Web:

    def test_cors(self, *, web):
        assert web.cors is web.cors
        assert isinstance(web.cors, aiohttp_cors.CorsConfig)

    def test_text(self, *, web):
        with patch('faust.web.drivers.aiohttp.Response') as Response:
            resp = web.text(
                'foo',
                content_type='app/json',
                status=303,
                reason='bar',
                headers={'k': 'v'},
            )
            Response.assert_called_once_with(
                content_type='app/json',
                status=303,
                text='foo',
                reason='bar',
                headers={'k': 'v'},
            )
            assert resp is Response()

    def test_json__from_str(self, *, web):
        web.text = Mock(name='web.text')
        with patch('faust.utils.json.dumps') as dumps:
            dumps.return_value = 'foo'
            payload = {'foo': 'bar'}
            response = web.json(
                payload,
                status=101,
                reason='bar',
                headers={'k': 'v'},
            )
            dumps.assert_called_once_with(payload)
            web.text.assert_called_once_with(
                dumps.return_value,
                content_type='application/json',
                status=101,
                reason='bar',
                headers={'k': 'v'},
            )
            assert response is web.text.return_value

    def test_json__from_bytes(self, *, web):
        web.bytes = Mock(name='web.bytes')
        with patch('faust.utils.json.dumps') as dumps:
            dumps.return_value = b'foo'
            payload = {'foo': 'bar'}
            response = web.json(
                payload,
                content_type='application/x-json',
                status=101,
                reason='bar',
                headers={'k': 'v'},
            )
            dumps.assert_called_once_with(payload)
            web.bytes.assert_called_once_with(
                dumps.return_value,
                content_type='application/x-json',
                status=101,
                reason='bar',
                headers={'k': 'v'},
            )
            assert response is web.bytes.return_value

    def test_html(self, *, web):
        with patch('faust.web.drivers.aiohttp.Response') as Response:
            resp = web.html(
                'foo',
                status=303,
                content_type='text/html',
                reason='bar',
                headers={'k': 'v'},
            )
            Response.assert_called_once_with(
                content_type='text/html',
                status=303,
                text='foo',
                reason='bar',
                headers={'k': 'v'},
            )
            assert resp is Response()

    def test_bytes(self, *, web):
        with patch('faust.web.drivers.aiohttp.Response') as Response:
            resp = web.bytes(
                b'foo',
                content_type='app/json',
                status=303,
                reason='bar',
                headers={'k': 'v'},
            )
            Response.assert_called_once_with(
                body=b'foo',
                content_type='app/json',
                status=303,
                reason='bar',
                headers={'k': 'v'},
            )
            assert resp is Response()

    @pytest.mark.asyncio
    async def test_on_start(self, *, web):
        web.add_dependency = Mock(name='add_dependency')
        web.app.conf.web_in_thread = True
        with patch('faust.web.drivers.aiohttp.ServerThread') as ServerThread:
            await web.on_start()
            ServerThread.assert_called_once_with(
                web, loop=web.loop, beacon=web.beacon)
            assert web._thread is ServerThread()
            web.add_dependency.assert_called_once_with(web._thread)

    @pytest.mark.asyncio
    async def test_wsgi(self, *, web):
        web.init_server = Mock(name='init_server')
        assert await web.wsgi() is web.web_app
        web.init_server.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_start_server(self, *, app, web):
        web.web_app = Mock(name='web.web_app', autospec=Application)
        web._runner = Mock(name='runner', setup=AsyncMock())
        web._create_site = Mock(
            name='create_site',
            return_value=Mock(
                start=AsyncMock(),
            ),
        )

        await web.start_server()

        web._runner.setup.assert_called_once_with()
        web._create_site.assert_called_once_with()
        web._create_site().start.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_stop_server(self, *, web):
        web._cleanup_app = AsyncMock(name='_cleanup_app')

        await web.stop_server()
        web._cleanup_app.assert_called_once_with()
        web._runner = None
        await web.stop_server()

    @pytest.mark.asyncio
    async def test_cleanup_app(self, *, web):
        web.web_app = None
        await web._cleanup_app()
        web.web_app = Mock(
            name='web.web_app',
            autospec=Application,
            cleanup=AsyncMock(),
        )
        await web._cleanup_app()
        web.web_app.cleanup.assert_called_once_with()

    def test_add_static(self, *, web):
        web.web_app = Mock(name='web.web_app', autospec=Application)
        web.add_static('/prefix/', Path(DATAPATH), kw1=3)
        web.web_app.router.add_static.assert_called_once_with(
            '/prefix/', str(Path(DATAPATH)), kw1=3,
        )

    def test_route__with_cors_options(self, *, web):
        handler = Mock()
        cors_options = {
            'http://example.com': ResourceOptions(
                allow_credentials=True,
                expose_headers='*',
                allow_headers='*',
                max_age=300,
                allow_methods='*',
            ),
        }
        web.web_app = Mock()
        web._cors = Mock()

        web.route(
            pattern='/foo/',
            handler=handler,
            cors_options=cors_options,
        )

        web.web_app.router.add_route.assert_has_calls([
            call(method, '/foo/', ANY)
            for method in NON_OPTIONS_METHODS
        ])
        web._cors.add.assert_has_calls([
            call(
                web.web_app.router.add_route(),
                _prepare_cors_options(cors_options))
            for _ in NON_OPTIONS_METHODS
        ])

    def test__create_site(self, *, web, app):
        web._new_transport = Mock()
        ret = web._create_site()
        web._new_transport.assert_called_once_with(
            app.conf.web_transport.scheme)
        assert ret is web._new_transport()

    def test__new_transport__tcp(self, *, web, app):
        app.conf.web_transport = URL('tcp://')
        with patch('faust.web.drivers.aiohttp.TCPSite') as TCPSite:
            ret = web._create_site()
            TCPSite.assert_called_once_with(
                web._runner,
                app.conf.web_bind,
                app.conf.web_port,
            )
            assert ret is TCPSite()

    def test__new_transport__unix(self, *, web, app):
        app.conf.web_transport = URL('unix:///etc/foobar')
        with patch('faust.web.drivers.aiohttp.UnixSite') as UnixSite:
            ret = web._create_site()
            UnixSite.assert_called_once_with(
                web._runner,
                '/etc/foobar',
            )
            assert ret is UnixSite()

    def test__app__compat_property(self, *, web):
        assert web._app is web.web_app



"""Storing objects in tables.

This is also used to store data structures such as sets/lists.

"""
import abc

from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    MutableMapping,
    Optional,
    Set,
    Type,
)

from mode import Service

from faust.stores.base import Store
from faust.streams import current_event
from faust.types import EventT, TP
from faust.types.stores import StoreT
from faust.types.tables import CollectionT

from .table import Table


class ChangeloggedObject:
    """A changelogged object in a :class:`ChangeloggedObjectManager` store."""

    manager: 'ChangeloggedObjectManager'

    def __init__(self, manager: 'ChangeloggedObjectManager', key: Any) -> None:
        self.manager = manager
        self.key = key
        self.__post_init__()

    def __post_init__(self) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    def sync_from_storage(self, value: Any) -> None:
        """Sync value from storage."""
        ...

    @abc.abstractmethod
    def as_stored_value(self) -> Any:
        """Return value as represented in storage."""
        ...

    @abc.abstractmethod
    def apply_changelog_event(self, operation: int, value: Any) -> None:
        """Apply event in changelog topic to local table state."""
        ...


class ChangeloggedObjectManager(Store):
    """Store of changelogged objects."""

    ValueType: ClassVar[Type[ChangeloggedObject]]

    table: Table
    data: MutableMapping

    _storage: Optional[StoreT] = None
    _dirty: Set

    def __init__(self, table: Table, **kwargs: Any) -> None:
        self.table = table
        self.table_name = self.table.name
        self.data = {}
        self._dirty = set()
        Service.__init__(self, **kwargs)

    def send_changelog_event(self,
                             key: Any,
                             operation: int,
                             value: Any) -> None:
        """Send changelog event to the tables changelog topic."""
        event = current_event()
        self._dirty.add(key)
        self.table._send_changelog(event, (operation, key), value)

    def __getitem__(self, key: Any) -> ChangeloggedObject:
        if key in self.data:
            return self.data[key]
        s = self.data[key] = self.ValueType(self, key)
        return s

    def __setitem__(self, key: Any, value: Any) -> None:
        raise NotImplementedError(f'{self._table_type_name}: cannot set key')

    def __delitem__(self, key: Any) -> None:
        raise NotImplementedError(f'{self._table_type_name}: cannot del key')

    @property
    def _table_type_name(self) -> str:
        return f'{type(self.table).__name__}'

    async def on_start(self) -> None:
        """Call when the changelogged object manager starts."""
        await self.add_runtime_dependency(self.storage)

    async def on_stop(self) -> None:
        """Call when the changelogged object manager stops."""
        self.flush_to_storage()

    def persisted_offset(self, tp: TP) -> Optional[int]:
        """Get the last persisted offset for changelog topic partition."""
        return self.storage.persisted_offset(tp)

    def set_persisted_offset(self, tp: TP, offset: int) -> None:
        """Set the last persisted offset for changelog topic partition."""
        self.storage.set_persisted_offset(tp, offset)

    async def on_rebalance(self,
                           table: CollectionT,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        """Call when cluster is rebalancing."""
        await self.storage.on_rebalance(
            table, assigned, revoked, newly_assigned)

    async def on_recovery_completed(self,
                                    active_tps: Set[TP],
                                    standby_tps: Set[TP]) -> None:
        """Call when table recovery is completed after rebalancing."""
        self.sync_from_storage()

    def sync_from_storage(self) -> None:
        """Sync set contents from storage."""
        for key, value in self.storage.items():
            self[key].sync_from_storage(value)

    def flush_to_storage(self) -> None:
        """Flush set contents to storage."""
        for key in self._dirty:
            self.storage[key] = self.data[key].as_stored_value()
        self._dirty.clear()

    @Service.task
    async def _periodic_flush(self) -> None:  # pragma: no cover
        async for sleep_time in self.itertimer(2.0, name='SetManager.flush'):
            await self.sleep(sleep_time)
            self.flush_to_storage()

    def reset_state(self) -> None:
        """Reset table local state."""
        # delegate to underlying RocksDB store.
        self.storage.reset_state()

    @property
    def storage(self) -> StoreT:
        """Return underlying storage used by this set table."""
        if self._storage is None:
            self._storage = self.table._new_store_by_url(
                self.table._store or self.table.app.conf.store)
        return self._storage

    def apply_changelog_batch(self,
                              batch: Iterable[EventT],
                              to_key: Callable[[Any], Any],
                              to_value: Callable[[Any], Any]) -> None:
        """Apply batch of changelog events to local state."""
        tp_offsets: Dict[TP, int] = {}
        for event in batch:
            tp, offset = event.message.tp, event.message.offset
            tp_offsets[tp] = (
                offset if tp not in tp_offsets
                else max(offset, tp_offsets[tp])
            )

            if event.key is None:
                raise RuntimeError('Changelog key cannot be None')

            operation, key = event.key
            key = to_key(key)
            value: Any = to_value(event.value)
            self[key].apply_changelog_event(operation, value)

        for tp, offset in tp_offsets.items():
            self.set_persisted_offset(tp, offset)



"""URL utilities - Working with URLs."""
from typing import List, Optional, Union
from yarl import URL


def urllist(arg: Union[URL, str, List[str], List[URL]], *,
            default_scheme: str = None) -> List[URL]:
    """Create list of URLs.

    You can pass in a comma-separated string, or an actual list
    and this will convert that into a list of :class:`yarl.URL` objects.
    """
    if not arg:
        raise ValueError('URL argument cannot be falsy')

    if isinstance(arg, URL):
        arg = [arg]
    elif isinstance(arg, str):
        arg = arg.split(';')

    scheme = URL(arg[0]).scheme or default_scheme
    return [_ensure_scheme(scheme, URL(u)) for u in arg]


def _ensure_scheme(default_scheme: Optional[str], url: Union[URL]) -> URL:
    """Ensure URL has a default scheme.

    An URL like "localhost" will be returned with the default scheme
    added, while an URL with existing scheme is returned unmodified.
    """
    if default_scheme and not url.scheme:
        if url.is_absolute():
            return url.with_scheme(default_scheme)
        else:
            return URL(f'{default_scheme}://{url}')
    return url



import pytest
from collections import Counter
from itertools import cycle
from pprint import pformat
from typing import NamedTuple
from faust.types import EventT, StreamT
from .helpers import AgentCase


class _param(NamedTuple):
    concurrency: int
    num_messages: int


@pytest.mark.asyncio
@pytest.mark.parametrize(_param._fields, [
    _param(10, 100),
    _param(10, 1000),
    _param(2, 10),
    _param(2, 100),
    _param(100, 1000),
    _param(1000, 1000),
])
async def test_agent_concurrency__duplicates(
        concurrency,
        num_messages,
        *,
        app,
        logging):
    await AgentConcurrencyCase.run_test(
        app=app,
        num_messages=num_messages,
        concurrency=concurrency,
    )


class AgentConcurrencyCase(AgentCase):
    name = 'test_agent_concurrency_duplicates'

    def on_init(self) -> None:
        assert self.concurrency > 0
        assert not self.concurrency % 2, 'concurrency is even'

        self.expected_index = cycle(range(self.concurrency))
        self.processed_by_concurrency_index = Counter()

    async def on_agent_event(
            self, stream: StreamT, event: EventT) -> None:
        cur_index = stream.concurrency_index

        # Test for: perfect round-robin, uniform distribution
        assert cur_index == next(self.expected_index)
        self.processed_by_concurrency_index[cur_index] += 1

    async def assert_success(self) -> None:
        await super().assert_success()
        self.log.info(
            '- Final coroutine distribution:\n%s',
            pformat(self.processed_by_concurrency_index.most_common()))
        max_ = None
        for _, total in self.processed_by_concurrency_index.most_common():
            if max_ is None:
                max_ = total
            assert total == max_



"""Web server driver registry."""
from typing import Type
from mode.utils.imports import FactoryMapping
from faust.web.base import Web

__all__ = ['by_name', 'by_url']

DRIVERS: FactoryMapping[Type[Web]] = FactoryMapping(
    aiohttp='faust.web.drivers.aiohttp:Web',
)
DRIVERS.include_setuptools_namespace('faust.web.drivers')
by_name = DRIVERS.by_name
by_url = DRIVERS.by_url



"""Cache backend registry."""
from typing import Type
from mode.utils.imports import FactoryMapping
from faust.types.web import CacheBackendT

__all__ = ['by_name', 'by_url']

BACKENDS: FactoryMapping[Type[CacheBackendT]] = FactoryMapping(
    memory='faust.web.cache.backends.memory:CacheBackend',
    redis='faust.web.cache.backends.redis:CacheBackend',
    rediscluster='faust.web.cache.backends.redis:CacheBackend',
)
BACKENDS.include_setuptools_namespace('faust.web.cache')
by_name = BACKENDS.by_name
by_url = BACKENDS.by_url



import pytest
from faust.fixups import fixups
from faust.fixups.django import Fixup
from mode.utils.mocks import Mock, mask_module, patch_module


class test_Fixup:

    @pytest.fixture()
    def fixup(self, *, app):
        return Fixup(app)

    def test_fixup_env_enabled(self, *, app, fixup, monkeypatch):
        monkeypatch.setenv('DJANGO_SETTINGS_MODULE', 'settings')
        with patch_module('django'):
            assert fixup.enabled()
            assert any(isinstance(f, Fixup) for f in fixups(app))

    def test_fixup_env_enabled_no_django(self, *, app, fixup, monkeypatch):
        monkeypatch.setenv('DJANGO_SETTINGS_MODULE', 'settings')
        with mask_module('django'):
            with pytest.warns(UserWarning):
                assert not fixup.enabled()
                assert not any(isinstance(f, Fixup) for f in fixups(app))

    def test_fixup_env_disabled(self, *, app, fixup, monkeypatch):
        monkeypatch.delenv('DJANGO_SETTINGS_MODULE', raising=False)
        assert not fixup.enabled()
        assert not any(isinstance(f, Fixup) for f in fixups(app))

    def test_wait_for_django_calls_django_setup(self, *, fixup, monkeypatch):
        with patch_module('django',
                          'django.conf',
                          'django.conf.settings'):
            import django
            from django.conf import settings
            settings.configure()
            settings.DEBUG = False
            fixup.wait_for_django()
            django.setup.assert_called_once_with()
            settings.DEBUG = True
            with pytest.warns(UserWarning):
                fixup.wait_for_django()

    def test_autodiscover_modules(self, *, fixup):
        with patch_module('django',
                          'django.conf',
                          'django.core',
                          'django.apps'):
            from django.apps import apps
            config1 = Mock()
            config1.name = 'config1'
            config2 = Mock()
            config2.name = 'config2'
            apps.get_app_configs.return_value = [config1, config2]

            assert fixup.autodiscover_modules() == ['config1', 'config2']

    def test_on_worker_init(self, *, fixup):
        with patch_module('django',
                          'django.conf',
                          'django.core',
                          'django.core.checks'):
            import django
            from django.conf import settings
            from django.core.checks import run_checks

            settings.DEBUG = False
            fixup.on_worker_init()
            django.setup.assert_called_once_with()
            run_checks.assert_called_once_with()

            settings.DEBUG = True
            with pytest.warns(UserWarning):
                fixup.on_worker_init()

    def test_apps(self, *, fixup):
        with patch_module('django', 'django.apps'):
            from django.apps import apps
            assert fixup.apps is apps

    def test_settings(self, *, fixup):
        with patch_module('django', 'django.conf'):
            from django.conf import settings
            assert fixup.settings is settings



import sys
import faust
sys.path.insert(0, '.')
from t.bench.base import Benchmark  # noqa

app = faust.App('bench-baseline')
topic = app.topic('bench-baseline', value_serializer='raw', value_type=int)
Benchmark(app, topic).install(__name__)



import faust
import pytest
from faust.tables.sets import (
    ChangeloggedSet,
    ChangeloggedSetManager,
    OPERATION_ADD,
    OPERATION_DISCARD,
    OPERATION_UPDATE,
    SetAction,
    SetManagerOperation,
    SetTableManager,
    SetWindowSet,
)
from mode.utils.mocks import AsyncMock, Mock, call


@pytest.fixture()
def key():
    return Mock(name='key')


@pytest.fixture()
def table():
    return Mock(name='table')


class test_SetWindowSet:

    @pytest.fixture()
    def wrapper(self):
        return Mock(name='wrapper')

    @pytest.fixture()
    def wset(self, *, key, table, wrapper):
        return SetWindowSet(key, table, wrapper)

    def test_add(self, *, wset):
        event = Mock(name='event')
        wset._apply_set_operation = Mock()
        wset.add('value', event=event)
        wset._apply_set_operation.assert_called_once_with(
            'add', 'value', event,
        )

    def test_discard(self, *, wset):
        event = Mock(name='event')
        wset._apply_set_operation = Mock()
        wset.discard('value', event=event)
        wset._apply_set_operation.assert_called_once_with(
            'discard', 'value', event,
        )

    def test__apply_set_operation(self, *, wset, key, table, wrapper):
        event = Mock(name='event')
        wrange1 = Mock(name='window_range1')
        wrange2 = Mock(name='window_range2')
        table._window_ranges.return_value = [wrange1, wrange2]

        wset._apply_set_operation('op', 'value', event)

        wset.wrapper.get_timestamp.assert_called_once_with(event)
        wrapper.on_set_key.assert_called_once_with(key, 'value')
        table._get_key.assert_has_calls([
            call((key, wrange1)),
            call().op('value'),
            call((key, wrange2)),
            call().op('value'),
        ])


class test_ChangeloggedSet:

    @pytest.fixture()
    def manager(self):
        return Mock(name='manager')

    @pytest.fixture()
    def cset(self, *, manager, key):
        return ChangeloggedSet(manager, key)

    def test_constructor(self, *, cset):
        assert isinstance(cset.data, set)

    def test_on_add(self, *, cset, manager, key):
        cset.add('value')
        manager.send_changelog_event.assert_called_once_with(
            key, OPERATION_ADD, 'value',
        )

    def test_on_discard(self, *, cset, manager, key):
        cset.data.add('value')
        cset.discard('value')
        manager.send_changelog_event.assert_called_once_with(
            key, OPERATION_DISCARD, 'value',
        )

    def test_on_change__diff(self, *, cset, manager, key):
        cset.data.update({1, 2, 3, 4})
        cset.difference_update({2, 3, 4, 5, 6})
        manager.send_changelog_event.assert_called_once_with(
            key, OPERATION_UPDATE, [set(), {2, 3, 4}],
        )

    def test_on_change__update(self, *, cset, manager, key):
        cset.data.update({1, 2, 3, 4})
        cset.update({2, 3, 4, 5, 6})
        manager.send_changelog_event.assert_called_once_with(
            key, OPERATION_UPDATE, [{5, 6}, set()],
        )

    def test_sync_from_storage(self, *, cset):
        cset.data.update({1, 2, 3})
        cset.sync_from_storage({2, 3, 4})
        assert cset.data == {2, 3, 4}

    def test_as_stored_value(self, *, cset):
        cset.data.update({1, 2, 3})
        assert cset.as_stored_value() == {1, 2, 3}

    def test_apply_changelog_event__ADD(self, *, cset):
        cset.data.update({2})
        cset.apply_changelog_event(OPERATION_ADD, 3)
        assert cset.data == {2, 3}

    def test_apply_changelog_event__DISCARD(self, *, cset):
        cset.data.update({2, 3})
        cset.apply_changelog_event(OPERATION_DISCARD, 2)
        assert cset.data == {3}

    def test_apply_changelog_event__UPDATE(self, *, cset):
        cset.data.update({5, 6, 7, 8, 9})
        added = {2, 3, 4}
        removed = {6, 7, 8}
        cset.apply_changelog_event(OPERATION_UPDATE, [added, removed])
        assert cset.data == {2, 3, 4, 5, 9}

    def test_apply_changelog_event__not_implemented(self, *, cset):
        with pytest.raises(NotImplementedError):
            cset.apply_changelog_event(0xffff, {1, 2, 3})


def test_ChangeloggedSetManager():
    assert ChangeloggedSetManager.ValueType is ChangeloggedSet


class test_SetTableManager:

    @pytest.fixture()
    def stable(self, *, app):
        return app.SetTable('name', start_manager=True)

    @pytest.fixture()
    def man(self, *, stable):
        return SetTableManager(stable)

    def test_constructor_enabled(self, *, app, stable):
        stable.start_manager = True
        app.agent = Mock(name='agent')
        man = SetTableManager(stable)
        assert man.enabled
        app.agent.assert_called_once_with(
            channel=man.topic,
            name='faust.SetTable.manager',
        )

    def test_constructor_disabled(self, *, app, stable):
        stable.start_manager = False
        app.agent = Mock(name='agent')
        man = SetTableManager(stable)
        assert man
        assert not man.enabled
        app.agent.assert_not_called()

    @pytest.mark.asyncio
    async def test_add(self, *, man):
        man._send_operation = AsyncMock()
        await man.add('key', 'member')
        man._send_operation.coro.assert_called_once_with(
            SetAction.ADD, 'key', ['member'],
        )

    @pytest.mark.asyncio
    async def test_discard(self, *, man):
        man._send_operation = AsyncMock()
        await man.discard('key', 'member')
        man._send_operation.coro.assert_called_once_with(
            SetAction.DISCARD, 'key', ['member'],
        )

    @pytest.mark.asyncio
    async def test_clear(self, *, man):
        man._send_operation = AsyncMock()
        await man.clear('key')
        man._send_operation.coro.assert_called_once_with(
            SetAction.CLEAR, 'key', [],
        )

    @pytest.mark.asyncio
    async def test_difference_update(self, *, man):
        man._send_operation = AsyncMock()
        await man.difference_update('key', ['v1', 'v2'])
        man._send_operation.coro.assert_called_once_with(
            SetAction.DISCARD, 'key', ['v1', 'v2'],
        )

    @pytest.mark.asyncio
    async def test_intersection_update(self, *, man):
        man._send_operation = AsyncMock()
        await man.intersection_update('key', ['v1', 'v2'])
        man._send_operation.coro.assert_called_once_with(
            SetAction.INTERSECTION, 'key', ['v1', 'v2'],
        )

    @pytest.mark.asyncio
    async def test_symmetric_difference_update(self, *, man):
        man._send_operation = AsyncMock()
        await man.symmetric_difference_update('key', ['v1', 'v2'])
        man._send_operation.coro.assert_called_once_with(
            SetAction.SYMDIFF, 'key', ['v1', 'v2'],
        )

    def test__update(self, *, man):
        man.set_table = {'a': Mock(name='a'), 'b': Mock(name='b')}
        man._update('a', ['v1'])
        man.set_table['a'].update.assert_called_once_with(['v1'])

    def test__difference_update(self, *, man):
        man.set_table = {'a': Mock(name='a'), 'b': Mock(name='b')}
        man._difference_update('a', ['v1'])
        man.set_table['a'].difference_update.assert_called_once_with(['v1'])

    def test__clear(self, *, man):
        man.set_table = {'a': Mock(name='a'), 'b': Mock(name='b')}
        man._clear('a', [])
        man.set_table['a'].clear.assert_called_once_with()

    def test__intersection_update(self, *, man):
        man.set_table = {'a': Mock(name='a'), 'b': Mock(name='b')}
        man._intersection_update('a', ['v1', 'v2', 'v3'])
        man.set_table['a'].intersection_update.assert_called_once_with(
            ['v1', 'v2', 'v3'],
        )

    def test__symmetric_difference_update(self, *, man):
        man.set_table = {'a': Mock(name='a'), 'b': Mock(name='b')}
        man._symmetric_difference_update('a', ['v1', 'v2', 'v3'])
        man.set_table['a'].symmetric_difference_update.assert_called_once_with(
            ['v1', 'v2', 'v3'],
        )

    @pytest.mark.asyncio
    async def test__send_operation__disabled(self, *, man):
        man.enabled = False
        with pytest.raises(RuntimeError):
            await man._send_operation(SetAction.ADD, 'k', 'v')

    @pytest.mark.asyncio
    async def test__send_operation__enabled(self, *, man):
        man.enabled = True
        man.topic.send = AsyncMock()
        await man._send_operation(SetAction.ADD, 'k', ['v'])
        man.topic.send.assert_called_once_with(
            key='k',
            value=SetManagerOperation(action=SetAction.ADD, members=['v']),
        )

    @pytest.mark.asyncio
    async def test__send_operation__enabled__iterator(self, *, man):
        man.enabled = True
        man.topic.send = AsyncMock()
        await man._send_operation(SetAction.ADD, 'k', iter(['a', 'b']))
        man.topic.send.assert_called_once_with(
            key='k',
            value=SetManagerOperation(
                action=SetAction.ADD,
                members=['a', 'b'],
            ),
        )

    @pytest.mark.asyncio
    @pytest.mark.allow_lingering_tasks(count=1)
    async def test__modify_set(self, *, man):
        stream = Mock()
        man.set_table = {
            'k1': Mock(name='k1'),
            'k2': Mock(name='k2'),
            'k3': Mock(name='k3'),
            'k4': Mock(name='k4'),
            'k5': Mock(name='k5'),
            'k6': Mock(name='k6'),
            'k7': Mock(name='k7'),
            'k8': Mock(name='k8'),
        }

        class X(faust.Record):
            x: int
            y: int

        unknown_set_op = SetManagerOperation(
            action=SetAction.ADD,
            members=['v4'],
        )
        unknown_set_op.action = 'UNKNOWN'

        async def stream_items():
            yield ('k1', SetManagerOperation(
                action=SetAction.ADD,
                members=['v'],
            ))
            yield ('k2', SetManagerOperation(
                action=SetAction.DISCARD,
                members=['v2'],
            ))
            yield ('k3', SetManagerOperation(
                action=SetAction.DISCARD,
                members=[X(10, 30).to_representation()],
            ))
            yield ('k4', unknown_set_op)
            yield ('k5', SetManagerOperation(
                action=SetAction.ADD,
                members=[
                    X(10, 30).to_representation(),
                    X(20, 40).to_representation(),
                    'v3',
                ],
            ))
            yield ('k6', SetManagerOperation(
                action=SetAction.INTERSECTION,
                members=[
                    X(10, 30).to_representation(),
                    X(20, 40).to_representation(),
                    'v3',
                ],
            ))
            yield ('k7', SetManagerOperation(
                action=SetAction.SYMDIFF,
                members=[
                    X(10, 30).to_representation(),
                    X(20, 40).to_representation(),
                    'v3',
                ],
            ))
            yield ('k8', SetManagerOperation(
                action=SetAction.CLEAR,
                members=[],
            ))

        stream.items.side_effect = stream_items

        await man._modify_set(stream)

        man.set_table['k1'].update.assert_called_with(['v'])
        man.set_table['k2'].difference_update.assert_called_with(['v2'])
        man.set_table['k3'].difference_update.assert_called_with([X(10, 30)])
        man.set_table['k5'].update.assert_called_with([
            X(10, 30),
            X(20, 40),
            'v3',
        ])
        man.set_table['k6'].intersection_update.assert_called_with([
            X(10, 30),
            X(20, 40),
            'v3',
        ])
        man.set_table['k7'].symmetric_difference_update.assert_called_with([
            X(10, 30),
            X(20, 40),
            'v3',
        ])
        man.set_table['k8'].clear.assert_called_once_with()


class test_SetTable:

    @pytest.fixture()
    def stable(self, *, app):
        return app.SetTable('name')

    def test_constructor__with_suffix(self, *, app):
        t1 = app.SetTable('name', manager_topic_suffix='-foo')
        assert t1.manager_topic_suffix == '-foo'
        assert t1.manager_topic_name == 'name-foo'

    def test_constructor__with_specific_name(self, *, app):
        t1 = app.SetTable('name', manager_topic_name='foo')
        assert t1.manager_topic_name == 'foo'

    @pytest.mark.asyncio
    async def test_on_start(self, *, stable):
        stable.changelog_topic.maybe_declare = AsyncMock()
        stable.add_runtime_dependency = AsyncMock()
        stable.start_manager = False
        await stable.on_start()

        stable.start_manager = True
        await stable.on_start()
        stable.add_runtime_dependency.assert_has_calls([
            call(stable.manager)], any_order=True)
        await stable.on_stop()

    def test__new_store(self, *, stable):
        store = stable._new_store()
        assert isinstance(store, ChangeloggedSetManager)
        assert store.table is stable

    def test__getitem__(self, *, stable):
        assert stable['foo'] is not None



from contextlib import contextmanager
from collections import deque
from datetime import datetime, timedelta, timezone
from statistics import median
import pytest
from mode import label
from mode.utils.mocks import ANY, AsyncMock, Mock, patch
from faust.livecheck import Case
from faust.livecheck.exceptions import SuiteFailed
from faust.livecheck.models import State, TestReport


class test_Case:

    @pytest.mark.parametrize('arg,value,expected', [
        ('active', False, False),
        ('probability', 3.33, 3.33),
        ('warn_stalled_after', 4.44, 4.44),
        ('test_expires', 5.55, timedelta(seconds=5.55)),
        ('frequency', 6.66, 6.66),
        ('realtime_logs', True, True),
        ('max_history', 3000, 3000),
        ('max_consecutive_failures', 3, 3),
        ('url_timeout_total', 7.77, 7.77),
        ('url_timeout_connect', 8.88, 8.88),
        ('url_error_retries', 6, 6),
        ('url_error_delay_min', 9.99, 9.99),
        ('url_error_delay_backoff', 10.10, 10.10),
        ('url_error_delay_max', 11.11, 11.11),
    ])
    def test_constructor(self, arg, value, expected, *, livecheck):
        kwargs = {'app': livecheck, 'name': 'n'}
        case = Case(**{arg: value}, **kwargs)
        assert getattr(case, arg) == expected

    @pytest.mark.asyncio
    async def test__sampler(self, *, case):
        case._sample = AsyncMock()
        case.sleep = AsyncMock()

        def on_sample():
            if case._sample.call_count == 3:
                case._stopped.set()
        case._sample.coro.side_effect = on_sample

        await case._sampler(case)

        assert case._sample.call_count == 3

    @pytest.mark.asyncio
    async def test__sample(self, *, case):
        await case._sample()

        case.frequency_history.extend(range(100))
        case.latency_history.extend(range(100, 200))
        case.runtime_history.extend(range(200, 300))

        await case._sample()

        assert case.frequency_avg == median(case.frequency_history)
        assert case.latency_avg == median(case.latency_history)
        assert case.runtime_avg == median(case.runtime_history)

    @pytest.mark.asyncio
    async def test_maybe_trigger(self, *, case):
        case.trigger = AsyncMock('trigger')

        with patch('faust.livecheck.case.uuid'):
            with patch('faust.livecheck.case.uniform') as uniform:
                uniform.return_value = 1.0
                async with case.maybe_trigger() as test:
                    assert test is None
                case.trigger.assert_not_called()

                uniform.return_value = 0.0
                async with case.maybe_trigger() as test:
                    assert test is case.trigger.coro.return_value
                    assert case.current_test is test

    @pytest.mark.asyncio
    async def test_run(self, *, case):
        with pytest.raises(NotImplementedError):
            await Case.run(case)

    @pytest.mark.asyncio
    async def test_trigger(self, *, case):
        case.app = Mock(pending_tests=Mock(send=AsyncMock()))
        t = await case.trigger('id1', 30, kw=2)
        assert t.id == 'id1'
        case.app.pending_tests.send.coro.assert_called_once_with(
            key='id1', value=t,
        )

    def test_now(self, *, case):
        assert isinstance(case._now(), datetime)
        assert case._now().tzinfo == timezone.utc

    @pytest.mark.asyncio
    async def test_resolve_signal(self, *, case):
        key = 'k'
        event = Mock(name='event')
        case.signals[event.signal_name] = Mock(resolve=AsyncMock())
        await case.resolve_signal(key, event)
        case.signals[event.signal_name].resolve.coro.assert_called_once_with(
            key, event,
        )

    @pytest.mark.asyncio
    async def test_execute(
            self, *,
            case, execution, current_execution_stack, frozen_monotonic):
        case.Runner = Mock(name='case.Runner')
        runner = case.Runner.return_value
        runner.execute = AsyncMock(name='runner.execute')
        await case.execute(execution)

        case.Runner.assert_called_once_with(
            case, execution, started=frozen_monotonic.return_value)
        runner.execute.coro.assert_called_once_with()
        current_execution_stack.push.assert_called_once_with(runner)

    @pytest.mark.asyncio
    @pytest.mark.parametrize('started,last_received,frequency', [
        (100.0, None, None),
        (100.0, 50.0, 10.0),
        (100.0, 50.0, None),
    ])
    async def test_on_test_start(self, started, last_received, frequency, *,
                                 case, runner):
        case.latency_history = deque([0.03] * case.max_history)
        case.frequency_history = deque([0.04] * case.max_history)
        runner.started = started
        case.frequency = frequency
        case.last_test_received = last_received

        await case.on_test_start(runner)

        if last_received:
            time_since = started - last_received
            if frequency:
                latency = time_since - frequency
                assert case.latency_history[-1] == latency
                assert len(case.latency_history) == case.max_history
            assert case.frequency_history[-1] == time_since
            assert len(case.frequency_history) == case.max_history

    @pytest.yield_fixture()
    def frozen_monotonic(self):
        with self._patch_monotonic() as monotonic:
            yield monotonic

    def _patch_monotonic(self):
        return patch('faust.livecheck.case.monotonic')

    @pytest.mark.asyncio
    async def test_on_test_skipped(self, *, case, runner, frozen_monotonic):
        await case.on_test_skipped(runner)
        assert case.last_test_received is frozen_monotonic.return_value

    @pytest.mark.asyncio
    async def test_on_test_failed(self, *, case, runner):
        case._set_test_error_state = AsyncMock()
        await case.on_test_failed(runner, KeyError())
        case._set_test_error_state.coro.assert_called_once_with(State.FAIL)

    @pytest.mark.asyncio
    async def test_on_test_error(self, *, case, runner):
        case._set_test_error_state = AsyncMock()
        await case.on_test_error(runner, KeyError())
        case._set_test_error_state.coro.assert_called_once_with(State.ERROR)

    @pytest.mark.asyncio
    async def test_on_test_timeout(self, *, case, runner):
        case._set_test_error_state = AsyncMock()
        await case.on_test_timeout(runner, KeyError())
        case._set_test_error_state.coro.assert_called_once_with(State.TIMEOUT)

    @pytest.mark.asyncio
    @pytest.mark.parametrize('state,failures,fail_suite', [
        (State.FAIL, 0, False),
        (State.FAIL, 9, True),
        (State.ERROR, 0, False),
        (State.ERROR, 9, True),
        (State.STALL, 0, False),
        (State.STALL, 9, True),
    ])
    async def test__set_test_error_state(
            self, state, failures, fail_suite, *, case):
        case.max_consecutive_failures = 10
        case.consecutive_failures = failures
        case.on_suite_fail = AsyncMock()
        await case._set_test_error_state(state)
        assert case.status == state
        assert case.consecutive_failures == failures + 1
        assert case.total_failures == 1
        assert case.total_by_state[state] == 1
        if fail_suite:
            case.on_suite_fail.coro.assert_called_once_with(ANY)

    @pytest.mark.asyncio
    @pytest.mark.parametrize('initial_state,ts,now,failed,expected_state', [
        (State.INIT, 300, 300.0, None, State.PASS),
        (State.PASS, 300, 300.0, 100.1, State.PASS),
        (State.FAIL, 300, 300.0, 299.0, State.FAIL),
        (State.FAIL, 100, 300.0, None, State.PASS),
        (State.FAIL, 101, 300.0, 100.0, State.PASS),
        (State.FAIL, 99, 300.0, 100.0, State.FAIL),
    ])
    async def test_on_suite_pass(
            self, initial_state, ts, now, failed, expected_state,
            *,
            case, runner, execution):
        assert runner.test is execution
        runner.test.timestamp = Mock()
        runner.test.timestamp.timestamp.return_value = ts
        case.runtime_history = deque([3.03] * case.max_history)
        runner.runtime = 300.0
        with self.seconds_since_last_fail(case, now=now, failed=failed):
            case.status = initial_state
            await case.on_test_pass(runner)
            assert case.status == expected_state
            assert len(case.runtime_history) == case.max_history
            assert case.runtime_history[-1] == 300.0

    @pytest.mark.asyncio
    async def test_post_report(self, *, case):
        report = Mock()
        case.app.post_report = AsyncMock()
        await case.post_report(report)
        case.app.post_report.assert_called_once_with(report)

    @pytest.mark.asyncio
    async def test__send_frequency__first_stop(self, *, case, loop):
        case.frequency = 0.1
        case.sleep = AsyncMock()
        with patch('mode.services.Timer') as ti:

            async def on_itertimer(*args, **kwargs):
                case._stopped.set()
                yield 0.1
                yield 0.2
                yield 0.3
                yield 0.4
            ti.side_effect = on_itertimer

            await case._send_frequency(case)

    @pytest.mark.asyncio
    async def test__send_frequency__no_frequency(self, *, case, loop):
        case.frequency = 0.0
        case.sleep = AsyncMock()
        case.make_fake_request = AsyncMock()
        with patch('mode.services.Timer') as ti:

            async def on_itertimer(*args, **kwargs):
                case._stopped.set()
                yield 0.1
                yield 0.2
                yield 0.3
                yield 0.4
            ti.side_effect = on_itertimer

            await case._send_frequency(case)
        case.make_fake_request.assert_not_called()

    @pytest.mark.asyncio
    async def test__send_frequency__last_stop(self, *, case):
        case.frequency = 0.1
        case.sleep = AsyncMock()
        case.app.is_leader = Mock(return_value=False)
        with patch('mode.services.Timer') as ti:

            async def on_itertimer(*args, **kwargs):
                for val in [0.1, 0.2, 0.3, 0.4]:
                    await case.sleep(val)
                    yield val

            ti.side_effect = on_itertimer

            async def on_sleep(secs, **kwargs):
                if case.sleep.call_count >= 2:
                    case._stopped.set()
            case.sleep.side_effect = on_sleep

            await case._send_frequency(case)

    @pytest.mark.asyncio
    async def test__send_frequency__no_frequency_None(self, *, case):
        case.frequency = None
        await case._send_frequency(case)

    @pytest.mark.asyncio
    async def test__send_frequency__timer_ends(self, *, case):
        case.sleep = AsyncMock()
        case.frequency = 10.0
        case.app.is_leader = Mock(return_value=False)
        with patch('mode.services.Timer') as ti:
            async def on_itertimer(*args, **kwargs):
                for val in [0.1, 0.2, 0.3, 0.4]:
                    yield val
            ti.side_effect = on_itertimer
            await case._send_frequency(case)

    @pytest.mark.asyncio
    async def test__send_frequency(self, *, case):
        case.frequency = 0.1
        case.make_fake_request = AsyncMock()
        case.sleep = AsyncMock()
        case.app.is_leader = Mock(return_value=False)

        def on_make_fake_request():
            if case.make_fake_request.call_count == 3:
                case._stopped.set()
        case.make_fake_request.coro.side_effect = on_make_fake_request

        def on_is_leader():
            if case.app.is_leader.call_count >= 2:
                return True
        case.app.is_leader.side_effect = on_is_leader

        await case._send_frequency(case)

        assert case.make_fake_request.call_count == 3

    @pytest.mark.asyncio
    async def test__check_frequency(self, *, case):
        with patch('mode.services.Timer') as ti:

            async def on_itertimer(*args, **kwargs):
                for val in [0.1, 0.2, 0.3, 0.4, 0.5]:
                    yield val
            ti.side_effect = on_itertimer
            case.sleep = AsyncMock()
            await case._check_frequency(case)

    @pytest.mark.asyncio
    async def test__check_frequency__last(self, *, case, frozen_monotonic):
        frozen_monotonic.return_value = 600.0
        case.warn_stalled_after = 10.0
        case.on_suite_fail = AsyncMock()
        with patch('mode.services.Timer') as ti:

            async def on_itertimer(*args, **kwargs):
                case.last_test_received = 10.0
                for val in [0.1, 0.2, 0.3, 0.4, 0.5]:
                    yield val
            ti.side_effect = on_itertimer
            case.sleep = AsyncMock()

            await case._check_frequency(case)
        case.on_suite_fail.assert_called_once_with(ANY, State.STALL)

    @pytest.mark.asyncio
    async def test__check_frequency__should_stop1(self, *, case):
        with patch('mode.services.Timer') as ti:

            async def on_itertimer(*args, **kwargs):
                case._stopped.set()
                yield 0.1
                yield 0.2
                yield 0.3
                yield 0.4
                yield 0.5
            ti.side_effect = on_itertimer

            case.sleep = AsyncMock()
            await case._check_frequency(case)

    @pytest.mark.asyncio
    async def test__check_frequency__last_stop(self, *, case):
        with patch('mode.services.Timer') as ti:
            case._stopped.clear()
            assert not case.should_stop

            async def on_itertimer(*args, **kwargs):
                for val in [0.1, 0.2, 0.3, 0.4, 0.5]:
                    await case.sleep(val)
                    yield val
            ti.side_effect = on_itertimer
            case.sleep = AsyncMock()

            async def on_sleep(arg, **kwargs):
                if case.sleep.call_count >= 2:
                    case._stopped.set()
            case.sleep.side_effect = on_sleep

            await case._check_frequency(case)

    @pytest.mark.asyncio
    @pytest.mark.parametrize('initial_state,now,failed,posts_report', [
        (State.PASS, 300.0, 100.0, True),
        (State.PASS, 300.0, 290.0, False),
        (State.FAIL, 300.0, 290.0, False),
        (State.FAIL, 300.0, 100.0, True),
        (State.TIMEOUT, 300.0, 290.0, False),
        (State.TIMEOUT, 300.0, 100.0, True),
    ])
    async def test_on_suite_fail(
            self, initial_state, now, failed, posts_report, *, case):
        case.log = Mock()
        case.post_report = AsyncMock()
        exc = None
        try:
            raise SuiteFailed('foo')
        except SuiteFailed as e:
            exc = e
        with self.seconds_since_last_fail(case, now=now, failed=failed):
            case.status = initial_state
            await case.on_suite_fail(exc)
            assert case.status == State.FAIL
            assert case.last_fail == now
            if posts_report:
                case.log.exception.assert_called_once_with(str(exc))
                case.post_report.coro.assert_called_once_with(TestReport(
                    case_name=case.name,
                    state=State.FAIL,
                    test=None,
                    runtime=None,
                    signal_latency={},
                    error=str(exc),
                    traceback=ANY,
                ))

    @pytest.mark.parametrize('initial_state,now,failed,expected_state', [
        (State.INIT, 300.0, None, State.PASS),
        (State.PASS, 300.0, 100.1, State.PASS),
        (State.FAIL, 300.0, 299.0, State.FAIL),
        (State.FAIL, 300.0, 100.0, State.PASS),
    ])
    def test__maybe_recover_from_failed_state(
            self, initial_state, now, failed, expected_state, *, case):
        with self.seconds_since_last_fail(case, now=now, failed=failed):
            case.status = initial_state
            case._maybe_recover_from_failed_state()
            assert case.status == expected_state

    @pytest.mark.parametrize('now,failed,arg,expected', [
        (300.0, 100.0, 10.0, True),
        (101.0, 100.0, 10.0, False),
        (300.0, None, 10.0, True),
    ])
    def test_failed_longer_than(self, now, failed, arg, expected, *, case):
        with self.seconds_since_last_fail(case, now=now, failed=failed):
            assert case._failed_longer_than(arg) == expected

    @pytest.mark.parametrize('now,failed,expected', (
        (300.0, 100.0, pytest.approx(200.0)),
        (300.0, None, None),
    ))
    def test_seconds_since_last_fail(self, now, failed, expected, *, case):
        with self.seconds_since_last_fail(case, now=now, failed=failed):
            if expected is None:
                assert case.seconds_since_last_fail is None
            else:
                assert case.seconds_since_last_fail == expected

    @contextmanager
    def seconds_since_last_fail(self, case, now, failed):
        with self._patch_monotonic() as monotonic:
            monotonic.return_value = now
            case.last_fail = failed
            yield

    @pytest.mark.asyncio
    async def test_get_url(self, *, case):
        url = 'http://foo/'
        await self.assert_url_called(case, case.get_url(url, kw=1),
                                     'get', url, kw=1)

    @pytest.mark.asyncio
    async def test_post_url(self, *, case):
        url = 'http://foo/'
        await self.assert_url_called(case, case.post_url(url, kw=1),
                                     'post', url, kw=1)

    async def assert_url_called(self, case, fut, method, url, **kwargs):
        case.url_request = AsyncMock('url_request')
        response = await fut
        assert response is case.url_request.coro.return_value
        case.url_request.coro.assert_called_once_with(
            method, url, **kwargs,
        )

    @pytest.mark.asyncio
    async def test_url_request(self, *, case, mock_http_client):
        case.app._http_client = mock_http_client
        case._maybe_recover_from_failed_state = Mock()
        await case.url_request('get', 'http://foo/') == 'foo'
        case._maybe_recover_from_failed_state.assert_called_once_with()

    @pytest.mark.asyncio
    @pytest.mark.http_session(status_code=500, max_failures=None)
    async def test_url_request_fails(self, *, case, mock_http_client):
        case.on_suite_fail = AsyncMock()
        case.sleep = AsyncMock()
        case.app._http_client = mock_http_client
        await case.url_request('get', 'http://foo/') is None
        case.on_suite_fail.assert_called_once_with(ANY)

    @pytest.mark.asyncio
    @pytest.mark.http_session(status_code=500, max_failures=3)
    async def test_url_request_fails_recover(self, *, case, mock_http_client):
        case.sleep = AsyncMock()
        case.app._http_client = mock_http_client
        await case.url_request('get', 'http://foo/') == 'foo'

    def test_current_test(self, *, case):
        with patch('faust.livecheck.case.current_test_stack') as cts:
            assert case.current_test is cts.top

    def test_current_execution(self, *, case):
        with patch('faust.livecheck.case.current_execution_stack') as ces:
            assert case.current_execution is ces.top

    def test_label(self, *, case):
        assert label(case) == case.label



from faust.fixups.base import Fixup


class test_Fixup:

    def test_init(self, *, app):
        assert Fixup(app).app is app

    def test_enabled(self, *, app):
        assert not Fixup(app).enabled()

    def test_autodiscover_modules(self, *, app):
        assert Fixup(app).autodiscover_modules() == []

    def test_on_worker_init(self, *, app):
        Fixup(app).on_worker_init()



"""Venusian (see :pypi:`venusian`).

We define our own interface so we don't have to specify the
callback argument.
"""
import venusian
from typing import Any, Callable
from venusian import Scanner, attach as _attach

__all__ = ['Scanner', 'attach']


def attach(fun: Callable, category: str, *,
           callback: Callable[[Scanner, str, Any], None] = None,
           **kwargs: Any) -> None:
    """Shortcut for :func:`venusian.attach`.

    This shortcut makes the callback argument optional.
    """
    callback = _on_found if callback is None else callback
    return _attach(fun, callback, category=category, **kwargs)


def _on_found(scanner: venusian.Scanner, name: str, obj: Any) -> None:
    ...



import pytz
from freezegun import freeze_time
from faust.utils.cron import secs_for_next

SECS_IN_HOUR = 60 * 60


@freeze_time('2000-01-01 00:00:00')
def test_secs_for_next():
    every_minute_cron_format = '*/1 * * * *'
    assert secs_for_next(every_minute_cron_format) == 60

    every_8pm_cron_format = '0 20 * * *'
    assert secs_for_next(every_8pm_cron_format) == 20 * SECS_IN_HOUR

    every_4th_july_1pm_cron_format = '0 13 4 7 *'
    days_until_4th_july = 31 + 28 + 31 + 30 + 31 + 30 + 4
    secs_until_4th_july = SECS_IN_HOUR * 24 * days_until_4th_july
    secs_until_1_pm = 13 * SECS_IN_HOUR
    total_secs = secs_until_4th_july + secs_until_1_pm
    assert secs_for_next(every_4th_july_1pm_cron_format) == total_secs


@freeze_time('2000-01-01 00:00:00')
def test_secs_for_next_with_tz():
    pacific = pytz.timezone('US/Pacific')

    every_8pm_cron_format = '0 20 * * *'
    # In Pacific time it's 16:00 so only 4 hours until 8:00pm
    assert secs_for_next(every_8pm_cron_format, tz=pacific) == 4 * SECS_IN_HOUR



"""Fixups - Base implementation."""
from typing import Iterable
from faust.types import AppT, FixupT

__all__ = ['Fixup']


class Fixup(FixupT):
    """Base class for fixups.

    Fixups are things that hook into Faust to make things
    work for other frameworks, such as Django.
    """

    def __init__(self, app: AppT) -> None:
        self.app = app

    def enabled(self) -> bool:
        """Return if fixup should be enabled in this environment."""
        return False

    def autodiscover_modules(self) -> Iterable[str]:
        """Return list of additional autodiscover modules."""
        return []

    def on_worker_init(self) -> None:
        """Call when initializing worker/CLI commands."""
        ...



import abc
import inspect
import logging
import socket
import ssl
import typing
import warnings
from datetime import timedelta, timezone, tzinfo
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    Union,
    cast,
)
from uuid import uuid4

from mode import Seconds, SupervisorStrategyT, want_seconds
from mode.utils.imports import SymbolArg, symbol_by_name
from mode.utils.logging import Severity
from yarl import URL

from faust.exceptions import AlreadyConfiguredWarning, ImproperlyConfigured
from faust.utils.urls import urllist

from ._env import DATADIR, WEB_BIND, WEB_PORT, WEB_TRANSPORT
from .agents import AgentT
from .assignor import LeaderAssignorT, PartitionAssignorT
from .auth import CredentialsArg, CredentialsT, to_credentials
from .codecs import CodecArg
from .enums import ProcessingGuarantee
from .events import EventT
from .router import RouterT
from .sensors import SensorT
from .serializers import RegistryT, SchemaT
from .streams import StreamT
from .transports import PartitionerT, SchedulingStrategyT
from .tables import GlobalTableT, TableManagerT, TableT
from .topics import TopicT
from .web import HttpClientT, ResourceOptions

if typing.TYPE_CHECKING:
    from .worker import Worker as _WorkerT
else:
    class _WorkerT: ...      # noqa

__all__ = ['AutodiscoverArg', 'Settings']

W_ALREADY_CONFIGURED = '''\
App is already configured.

Reconfiguring late may mean parts of your program are still
using the old configuration.

Code such as:

app.conf.config_from_object('my.config')

Should appear before calling app.topic/@app.agent/etc.
'''

W_ALREADY_CONFIGURED_KEY = '''\
Setting new value for configuration key {key!r} that was already used.

Reconfiguring late may mean parts of your program are still
using the old value of {old_value!r}.

Code such as:

app.conf.{key} = {value!r}

Should appear before calling app.topic/@app.agent/etc.
'''


# XXX mypy borks if we do `from faust import __version__`
faust_version: str = symbol_by_name('faust:__version__')

TIMEZONE = timezone.utc

#: Broker URL, used as default for :setting:`broker`.
BROKER_URL = 'kafka://localhost:9092'

#: Default transport used when no scheme specified.
DEFAULT_BROKER_SCHEME = 'kafka'

#: Table storage URL, used as default for :setting:`store`.
STORE_URL = 'memory://'

#: Cache storage URL, used as default for :setting:`cache`.
CACHE_URL = 'memory://'

#: Web driver URL, used as default for :setting:`web`.
WEB_URL = 'aiohttp://'

PROCESSING_GUARANTEE = ProcessingGuarantee.AT_LEAST_ONCE

#: Table state directory path used as default for :setting:`tabledir`.
#: This path will be treated as relative to datadir, unless the provided
#: poth is absolute.
TABLEDIR = 'tables'

#: Path to agent class, used as default for :setting:`Agent`.
AGENT_TYPE = 'faust.Agent'

#: Default agent supervisor type, used as default for
#: :setting:`agent_supervisor`.
AGENT_SUPERVISOR_TYPE = 'mode.OneForOneSupervisor'

CONSUMER_SCHEDULER_TYPE = 'faust.transport.utils.DefaultSchedulingStrategy'

#: Default event class type, used as default for :setting:`Event`.
EVENT_TYPE = 'faust.Event'

#: Default schema class type, used as default for :setting:`Schema`.
SCHEMA_TYPE = 'faust.Schema'

#: Path to stream class, used as default for :setting:`Stream`.
STREAM_TYPE = 'faust.Stream'

#: Path to table manager class, used as default for :setting:`TableManager`.
TABLE_MANAGER_TYPE = 'faust.tables.TableManager'

#: Path to table class, used as default for :setting:`Table`.
TABLE_TYPE = 'faust.Table'

#: Tables keep a cache of key to partition number. This value
#: configures the maximum size of that cache.
TABLE_KEY_INDEX_SIZE = 1000

#: Path to "table of sets" class, used as default for :setting:`SetTable`.
SET_TABLE_TYPE = 'faust.SetTable'

#: Path to global table class, used as default for :setting:`GlobalTable`.
GLOBAL_TABLE_TYPE = 'faust.GlobalTable'

#: Path to "global table of sets" class,
#: used as default for :setting:`SetGlobalTable`.
SET_GLOBAL_TABLE_TYPE = 'faust.SetGlobalTable'

#: Path to serializer registry class, used as the default for
#: :setting:`Serializers`.
REGISTRY_TYPE = 'faust.serializers.Registry'

#: Path to worker class, providing the default for :setting:`Worker`.
WORKER_TYPE = 'faust.worker.Worker'

#: Path to partition assignor class, providing the default for
#: :setting:`PartitionAssignor`.
PARTITION_ASSIGNOR_TYPE = 'faust.assignor:PartitionAssignor'

#: Path to leader assignor class, providing the default for
#: :setting:`LeaderAssignor`.
LEADER_ASSIGNOR_TYPE = 'faust.assignor:LeaderAssignor'

#: Path to router class, providing the default for :setting:`Router`.
ROUTER_TYPE = 'faust.app.router:Router'

#: Path to topic class, providing the default for :setting:`Topic`.
TOPIC_TYPE = 'faust:Topic'

#: Path to HTTP client class, providing the default for :setting:`HttpClient`.
HTTP_CLIENT_TYPE = 'aiohttp.client:ClientSession'

#: Path to Monitor sensor class, providing the default for :setting:`Monitor`.
MONITOR_TYPE = 'faust.sensors:Monitor'

#: Default broker API version.
#: Used as default for
#:     + :setting:`broker_api_version`,
#:     + :setting:`consumer_api_version`,
#:     + :setting:`producer_api_version',
BROKER_API_VERSION = 'auto'

#: Default Kafka Client ID.
BROKER_CLIENT_ID = f'faust-{faust_version}'

#: Kafka consumer request timeout (``request_timeout_ms``).
BROKER_REQUEST_TIMEOUT = 90.0

#: How often we commit acknowledged messages: every n messages.
#: Used as the default value for :setting:`broker_commit_every`.
BROKER_COMMIT_EVERY = 10_000

#: How often we commit acknowledged messages on a timer.
#: Used as the default value for :setting:`broker_commit_interval`.
BROKER_COMMIT_INTERVAL = 2.8

#: Kafka consumer session timeout (``session_timeout_ms`` in seconds).
BROKER_SESSION_TIMEOUT = 60.0

#: Kafka consumer heartbeat timeout (``rebalance_timeout_ms`` in seconds).
BROKER_REBALANCE_TIMEOUT = 60.0

#: Kafka consumer heartbeat (``heartbeat_interval_ms``).
BROKER_HEARTBEAT_INTERVAL = 3.0

#: How long time it takes before we warn that the commit offset has
#: not advanced.
BROKER_LIVELOCK_SOFT = want_seconds(timedelta(minutes=5))

#: The maximum number of records returned in a single call to poll().
#: If you find that your application needs more time to process messages
#: you may want to adjust max_poll_records to tune the number of records
#: that must be handled on every loop iteration.
BROKER_MAX_POLL_RECORDS = None

#: Maximum allowed time between calls to consume messages
#: (e.g., consumer.getmany()).If this interval is exceeded the consumer
#: is considered failed and the group will rebalance in order to reassign
#: the partitions to another consumer group member. If API methods block
#: waiting for messages, that time does not count against this timeout.
#: See KIP-62 for more information.
BROKER_MAX_POLL_INTERVAL = 1000.0

#: How often we clean up expired items in windowed tables.
#: Used as the default value for :setting:`table_cleanup_interval`.
TABLE_CLEANUP_INTERVAL = 30.0

#: Prefix used for reply topics.
REPLY_TO_PREFIX = 'f-reply-'

#: Default expiry time for replies, in seconds (float).
REPLY_EXPIRES = want_seconds(timedelta(days=1))

#: Max number of messages channels/streams/topics can "prefetch".
STREAM_BUFFER_MAXSIZE = 4096

#: Number of seconds to sleep before continuing after rebalance.
#: We wait for a bit to allow for more nodes to join/leave before
#: starting to reprocess, to minimize the chance of errors and rebalancing
#: loops.
STREAM_RECOVERY_DELAY = 0.0

#: We buffer up sending messages until the
#: source topic offset related to that processing is committed.
#: This means when we do commit, we may have buffered up a LOT of messages
#: so commit frequently.
#:
#: This setting is deprecated and will be removed once transaction support
#: is added in a later version.
STREAM_PUBLISH_ON_COMMIT = False

#: Timeout (in seconds) for processing events in the stream.
#: If processing of a single event exceeds this time we log an error,
#: but do not stop processing.
#:
#: If you are seeing a warning like this you should either
#: 1) increase this timeout to allow events to take longer, or
#: 2) add a timeout to the operation so that stream processing
#:    always completed before the timeout.
#:
#: The latter is preferred for network operations such as web requests.
#: If a network service you depend on is temporarily offline you should
#: consider doing retries (sent to a separate topic).
#:
#:
#: main_topic = app.topic('main')
#: deadletter_topic = app.topic('main_deadletter')
#:
#: async def send_request(value, timeout: float = None) -> None:
#:     await app.http_client.get('http://foo.com', timeout=timeout)
#:
#: @app.agent(main_topic)
#: async def main(stream):
#:    async for value in stream:
#:        try:
#:            await send_request(value, timeout=5)
#:        except asyncio.TimeoutError:
#:            await deadletter_topic.send(value)
#:
#:
#: @app.agent(deadletter_topic)
#: async def main_deadletter(stream):
#:     async for value in stream:
#:         # wait for 30 seconds before retrying.
#:         await stream.sleep(30)
#:         await send_request(value)
STREAM_PROCESSING_TIMEOUT = 5 * 60.0

#: Maximum size of a request in bytes in the consumer.
#: Used as the default value for :setting:`max_fetch_size`.
#: Note: This is PER PARTITION, so a limit of 1Mb when your
#: workers consume from 10 topics having 100 partitions each
#: means a fetch request can be 10 * 100 * 1Mb.
#: This limit being effectively too high will cause
#: rebalancing issues, so keep it low enough to account
#: for many partitions.
CONSUMER_MAX_FETCH_SIZE = 1024 ** 2

#: Where the consumer should start reading offsets when there is no initial
#: offset, or the stored offset no longer exists, e.g. when starting a new
#: consumer for the first time. Options include 'earliest', 'latest', 'none'.
#: Used as default value for :setting:`consumer_auto_offset_reset`.
CONSUMER_AUTO_OFFSET_RESET = 'earliest'

#: Minimum time to batch before sending out messages from the producer.
#: Used as the default value for :setting:`linger_ms`.
PRODUCER_LINGER_MS = 0

#: Maximum size of buffered data per partition in bytes in the producer.
#: Used as the default value for :setting:`max_batch_size`.
PRODUCER_MAX_BATCH_SIZE = 16384

#: Maximum size of a request in bytes in the producer.
#: Used as the default value for :setting:`max_request_size`.
PRODUCER_MAX_REQUEST_SIZE = 1_000_000

#: The compression type for all data generated by
#: the producer. Valid values are 'gzip', 'snappy', 'lz4', or None.
#: Compression is of full batches of data, so the efficacy of batching
#: will also impact the compression ratio (more batching means better
#: compression). Default: None.
PRODUCER_COMPRESSION_TYPE: Optional[str] = None

#: Producer request timeout is the number of seconds before we give
#: up sending message batches.
PRODUCER_REQUEST_TIMEOUT: float = 1200.0  # 20 minutes.

#: The number of acknowledgments the producer requires the leader to have
#: received before considering a request complete. This controls the
#: durability of records that are sent. The following settings are common:
#:     0: Producer will not wait for any acknowledgment from the server
#:         at all. The message will immediately be considered sent.
#:         (Not recommended)
#:     1: The broker leader will write the record to its local log but
#:         will respond without awaiting full acknowledgment from all
#:         followers. In this case should the leader fail immediately
#:         after acknowledging the record but before the followers have
#:         replicated it then the record will be lost.
#:     -1: The broker leader will wait for the full set of in-sync
#:         replicas to acknowledge the record. This guarantees that the
#:         record will not be lost as long as at least one in-sync replica
#:         remains alive. This is the strongest available guarantee.
PRODUCER_ACKS = -1

#: Set of settings added for backwards compatibility
SETTINGS_COMPAT: Set[str] = {
    'stream_ack_exceptions',
    'stream_ack_cancelled_tasks',
    'url',
}

#: Set of :func:`inspect.signature` parameter types to ignore
#: when building the list of supported seting names.
SETTINGS_SKIP: Set[inspect._ParameterKind] = {
    inspect.Parameter.VAR_KEYWORD,
}

#: Memory utilization threshold that triggers a faust worker to
#: shutdown. Applicable only when faust is deployed using a `Worker`.
#:
#: An app will abruptly terminate if it exceeds it's allocated memory limit.
#: This may have unintendent consequences, especially when sateful Tables are
#: involved. This settings allows for graceful shutdown when high memory utilization
#: is detected. This setting is disabled by default. To enable, set percent value
#: between >0.0 and <=1.0
WORKER_SHUTDOWN_MEMORY_UTILIZATION_PERCENT = 0.0

#: Memory utilization (bytes) threshold that triggers a faust worker to
#: shutdown. Applicable only when faust is deployed using a `Worker`.
WORKER_SHUTDOWN_MEMORY_UTILIZATION_BYTES = 0.0

AutodiscoverArg = Union[
    bool,
    Iterable[str],
    Callable[[], Iterable[str]],
]


class Settings(abc.ABC):
    debug: bool = False
    autodiscover: AutodiscoverArg = False
    broker_client_id: str = BROKER_CLIENT_ID
    broker_api_version: str = BROKER_API_VERSION
    broker_commit_every: int = BROKER_COMMIT_EVERY
    broker_check_crcs: bool = True
    broker_max_poll_interval: float = BROKER_MAX_POLL_INTERVAL
    _broker_credentials: Optional[CredentialsT] = None
    id_format: str = '{id}-v{self.version}'
    key_serializer: CodecArg = 'raw'
    value_serializer: CodecArg = 'json'
    reply_to: str
    reply_to_prefix: str = REPLY_TO_PREFIX
    reply_create_topic: bool = False
    stream_buffer_maxsize: int = STREAM_BUFFER_MAXSIZE
    stream_wait_empty: bool = True
    stream_ack_cancelled_tasks: bool = True
    stream_ack_exceptions: bool = True
    stream_publish_on_commit: bool = STREAM_PUBLISH_ON_COMMIT
    ssl_context: Optional[ssl.SSLContext] = None
    table_standby_replicas: int = 1
    table_key_index_size: int = TABLE_KEY_INDEX_SIZE
    topic_replication_factor: int = 1
    topic_partitions: int = 8  # noqa: E704
    topic_allow_declare: bool = True
    topic_disable_leader: bool = False
    logging_config: Optional[Dict] = None
    loghandlers: List[logging.Handler]
    producer_linger_ms: int = PRODUCER_LINGER_MS
    producer_max_batch_size: int = PRODUCER_MAX_BATCH_SIZE
    producer_acks: int = PRODUCER_ACKS
    producer_max_request_size: int = PRODUCER_MAX_REQUEST_SIZE
    consumer_group_instance_id: Optional[str] = None
    consumer_max_fetch_size: int = CONSUMER_MAX_FETCH_SIZE
    consumer_auto_offset_reset: str = CONSUMER_AUTO_OFFSET_RESET
    producer_compression_type: Optional[str] = PRODUCER_COMPRESSION_TYPE
    timezone: tzinfo = TIMEZONE
    web_enabled: bool
    web_bind: str = WEB_BIND
    web_port: int = WEB_PORT
    web_host: str = socket.gethostname()
    web_in_thread: bool = False
    web_cors_options: Optional[Mapping[str, ResourceOptions]] = None
    worker_redirect_stdouts: bool = True
    worker_redirect_stdouts_level: Severity = 'WARN'
    worker_shutdown_memory_utilization_percent: float = WORKER_SHUTDOWN_MEMORY_UTILIZATION_PERCENT
    worker_shutdown_memory_utilization_bytes: float = WORKER_SHUTDOWN_MEMORY_UTILIZATION_BYTES

    _id: str
    _origin: Optional[str] = None
    _name: str
    _version: int = 1
    _broker: List[URL]
    _broker_consumer: Optional[List[URL]] = None
    _broker_producer: Optional[List[URL]] = None
    _store: URL
    _cache: URL
    _web: URL
    _canonical_url: URL = cast(URL, None)
    _datadir: Path
    _tabledir: Path
    _processing_guarantee: ProcessingGuarantee = PROCESSING_GUARANTEE
    _agent_supervisor: Type[SupervisorStrategyT]
    _broker_request_timeout: float = BROKER_REQUEST_TIMEOUT
    _broker_session_timeout: float = BROKER_SESSION_TIMEOUT
    _broker_rebalance_timeout: float = BROKER_REBALANCE_TIMEOUT
    _broker_heartbeat_interval: float = BROKER_HEARTBEAT_INTERVAL
    _broker_commit_interval: float = BROKER_COMMIT_INTERVAL
    _broker_commit_livelock_soft_timeout: float = BROKER_LIVELOCK_SOFT
    _broker_max_poll_records: Optional[int] = BROKER_MAX_POLL_RECORDS
    _consumer_api_version: Optional[str] = None
    _producer_api_version: Optional[str] = None
    _producer_partitioner: Optional[PartitionerT] = None
    _producer_request_timeout: float = PRODUCER_REQUEST_TIMEOUT
    _stream_recovery_delay: float = STREAM_RECOVERY_DELAY
    _stream_processing_timeout: float = STREAM_PROCESSING_TIMEOUT
    _table_cleanup_interval: float = TABLE_CLEANUP_INTERVAL
    _reply_expires: float = REPLY_EXPIRES
    _web_transport: URL = WEB_TRANSPORT
    _Agent: Type[AgentT]
    _ConsumerScheduler: Type[SchedulingStrategyT]
    _Event: Type[EventT]
    _Schema: Type[SchemaT]
    _Stream: Type[StreamT]
    _Table: Type[TableT]
    _SetTable: Type[TableT]
    _GlobalTable: Type[GlobalTableT]
    _SetGlobalTable: Type[GlobalTableT]
    _TableManager: Type[TableManagerT]
    _Serializers: Type[RegistryT]
    _Worker: Type[_WorkerT]
    _PartitionAssignor: Type[PartitionAssignorT]
    _LeaderAssignor: Type[LeaderAssignorT]
    _Router: Type[RouterT]
    _Topic: Type[TopicT]
    _HttpClient: Type[HttpClientT]
    _Monitor: Type[SensorT]

    _initializing: bool = True
    _accessed: Set[str] = cast(Set[str], None)

    @classmethod
    def setting_names(cls) -> Set[str]:
        return {
            k for k, v in inspect.signature(cls).parameters.items()
            if v.kind not in SETTINGS_SKIP
        } - SETTINGS_COMPAT

    @classmethod
    def _warn_already_configured(cls) -> None:
        warnings.warn(AlreadyConfiguredWarning(W_ALREADY_CONFIGURED),
                      stacklevel=3)

    @classmethod
    def _warn_already_configured_key(cls,
                                     key: str,
                                     value: Any,
                                     old_value: Any) -> None:
        warnings.warn(
            AlreadyConfiguredWarning(W_ALREADY_CONFIGURED_KEY.format(
                key=key,
                value=value,
                old_value=old_value,
            )),
            stacklevel=3,
        )

    def __init__(  # noqa: C901
            self,
            id: str,
            *,
            debug: bool = None,
            version: int = None,
            broker: Union[str, URL, List[URL]] = None,
            broker_api_version: str = None,
            broker_client_id: str = None,
            broker_request_timeout: Seconds = None,
            broker_credentials: CredentialsArg = None,
            broker_commit_every: int = None,
            broker_commit_interval: Seconds = None,
            broker_commit_livelock_soft_timeout: Seconds = None,
            broker_session_timeout: Seconds = None,
            broker_rebalance_timeout: Seconds = None,
            broker_heartbeat_interval: Seconds = None,
            broker_check_crcs: bool = None,
            broker_max_poll_records: int = None,
            broker_max_poll_interval: int = None,
            broker_consumer: Union[str, URL, List[URL]] = None,
            broker_producer: Union[str, URL, List[URL]] = None,
            agent_supervisor: SymbolArg[Type[SupervisorStrategyT]] = None,
            store: Union[str, URL] = None,
            cache: Union[str, URL] = None,
            web: Union[str, URL] = None,
            web_enabled: bool = True,
            processing_guarantee: Union[str, ProcessingGuarantee] = None,
            timezone: tzinfo = None,
            autodiscover: AutodiscoverArg = None,
            origin: str = None,
            canonical_url: Union[str, URL] = None,
            datadir: Union[Path, str] = None,
            tabledir: Union[Path, str] = None,
            key_serializer: CodecArg = None,
            value_serializer: CodecArg = None,
            logging_config: Dict = None,
            loghandlers: List[logging.Handler] = None,
            table_cleanup_interval: Seconds = None,
            table_standby_replicas: int = None,
            table_key_index_size: int = None,
            topic_replication_factor: int = None,
            topic_partitions: int = None,
            topic_allow_declare: bool = None,
            topic_disable_leader: bool = None,
            id_format: str = None,
            reply_to: str = None,
            reply_to_prefix: str = None,
            reply_create_topic: bool = None,
            reply_expires: Seconds = None,
            ssl_context: ssl.SSLContext = None,
            stream_buffer_maxsize: int = None,
            stream_wait_empty: bool = None,
            stream_ack_cancelled_tasks: bool = None,
            stream_ack_exceptions: bool = None,
            stream_publish_on_commit: bool = None,
            stream_recovery_delay: Seconds = None,
            stream_processing_timeout: Seconds = None,
            producer_linger_ms: int = None,
            producer_max_batch_size: int = None,
            producer_acks: int = None,
            producer_max_request_size: int = None,
            producer_compression_type: str = None,
            producer_partitioner: SymbolArg[PartitionerT] = None,
            producer_request_timeout: Seconds = None,
            producer_api_version: str = None,
            consumer_group_instance_id: str = None,
            consumer_api_version: str = None,
            consumer_max_fetch_size: int = None,
            consumer_auto_offset_reset: str = None,
            web_bind: str = None,
            web_port: int = None,
            web_host: str = None,
            web_transport: Union[str, URL] = None,
            web_in_thread: bool = None,
            web_cors_options: Mapping[str, ResourceOptions] = None,
            worker_redirect_stdouts: bool = None,
            worker_redirect_stdouts_level: Severity = None,
            worker_shutdown_memory_utilization_percent: float = None,
            worker_shutdown_memory_utilization_bytes: float = None,
            Agent: SymbolArg[Type[AgentT]] = None,
            ConsumerScheduler: SymbolArg[Type[SchedulingStrategyT]] = None,
            Event: SymbolArg[Type[EventT]] = None,
            Schema: SymbolArg[Type[SchemaT]] = None,
            Stream: SymbolArg[Type[StreamT]] = None,
            Table: SymbolArg[Type[TableT]] = None,
            SetTable: SymbolArg[Type[TableT]] = None,
            GlobalTable: SymbolArg[Type[GlobalTableT]] = None,
            SetGlobalTable: SymbolArg[Type[GlobalTableT]] = None,
            TableManager: SymbolArg[Type[TableManagerT]] = None,
            Serializers: SymbolArg[Type[RegistryT]] = None,
            Worker: SymbolArg[Type[_WorkerT]] = None,
            PartitionAssignor: SymbolArg[Type[PartitionAssignorT]] = None,
            LeaderAssignor: SymbolArg[Type[LeaderAssignorT]] = None,
            Router: SymbolArg[Type[RouterT]] = None,
            Topic: SymbolArg[Type[TopicT]] = None,
            HttpClient: SymbolArg[Type[HttpClientT]] = None,
            Monitor: SymbolArg[Type[SensorT]] = None,
            # XXX backward compat (remove for Faust 1.0)
            url: Union[str, URL] = None,
            **kwargs: Any) -> None:
        object.__setattr__(self, '_accessed', None)
        self.version = version if version is not None else self._version
        self.id_format = id_format if id_format is not None else self.id_format
        if origin is not None:
            self.origin = origin
        self.id = id
        self.broker = self._first_not_none(url, broker, BROKER_URL)
        if debug is not None:
            self.debug = debug
        if broker_consumer is not None:
            # XXX have to cast to silence mypy bug
            self.broker_consumer = cast(List[URL], broker_consumer)
        if broker_producer is not None:
            # XXX have to cast to silence mypy bug
            self.broker_producer = cast(List[URL], broker_producer)
        self.ssl_context = ssl_context
        self.store = self._first_not_none(store, STORE_URL)
        self.cache = self._first_not_none(cache, CACHE_URL)
        self.web = self._first_not_none(web, WEB_URL)
        self.web_enabled = web_enabled
        if processing_guarantee is not None:
            # XXX have to cast to silence mypy bug
            self.processing_guarantee = cast(
                ProcessingGuarantee, processing_guarantee)
        if autodiscover is not None:
            self.autodiscover = autodiscover
        if broker_api_version is not None:
            self.broker_api_version = broker_api_version
        if broker_client_id is not None:
            self.broker_client_id = broker_client_id
        if canonical_url:
            # XXX have to cast to silence mypy bug
            self.canonical_url = cast(URL, canonical_url)
        # datadir is a format string that can contain e.g. {conf.id}
        # XXX have to cast to silence mypy bug
        self.datadir = cast(Path, datadir or DATADIR)
        # XXX have to cast to silence mypy bug
        self.tabledir = cast(Path, tabledir or TABLEDIR)
        if broker_credentials is not None:
            # XXX have to cast to silence mypy bug
            self.broker_credentials = cast(CredentialsT, broker_credentials)
        if broker_request_timeout is not None:
            self.broker_request_timeout = want_seconds(broker_request_timeout)
        self.broker_commit_interval = cast(float, (
            broker_commit_interval or self._broker_commit_interval))
        self.broker_commit_livelock_soft_timeout = cast(float, (
            broker_commit_livelock_soft_timeout or
            self._broker_commit_livelock_soft_timeout))
        if broker_session_timeout is not None:
            self.broker_session_timeout = want_seconds(broker_session_timeout)
        if broker_rebalance_timeout is not None:
            self.broker_rebalance_timeout = want_seconds(
                broker_rebalance_timeout)
        if broker_heartbeat_interval is not None:
            self.broker_heartbeat_interval = want_seconds(
                broker_heartbeat_interval)
        self.table_cleanup_interval = cast(float, (
            table_cleanup_interval or self._table_cleanup_interval))

        if timezone is not None:
            self.timezone = timezone

        if broker_commit_every is not None:
            self.broker_commit_every = broker_commit_every
        if broker_check_crcs is not None:
            self.broker_check_crcs = broker_check_crcs
        if broker_max_poll_records is not None:
            self.broker_max_poll_records = broker_max_poll_records
        if broker_max_poll_interval is not None:
            self.broker_max_poll_interval = broker_max_poll_interval
        if key_serializer is not None:
            self.key_serializer = key_serializer
        if value_serializer is not None:
            self.value_serializer = value_serializer
        if table_standby_replicas is not None:
            self.table_standby_replicas = table_standby_replicas
        if table_key_index_size is not None:
            self.table_key_index_size = table_key_index_size
        if topic_replication_factor is not None:
            self.topic_replication_factor = topic_replication_factor
        if topic_partitions is not None:
            self.topic_partitions = topic_partitions
        if topic_allow_declare is not None:
            self.topic_allow_declare = topic_allow_declare
        if topic_disable_leader is not None:
            self.topic_disable_leader = topic_disable_leader
        if reply_create_topic is not None:
            self.reply_create_topic = reply_create_topic
        if logging_config is not None:
            self.logging_config = logging_config
        self.loghandlers = loghandlers if loghandlers is not None else []
        if stream_buffer_maxsize is not None:
            self.stream_buffer_maxsize = stream_buffer_maxsize
        if stream_wait_empty is not None:
            self.stream_wait_empty = stream_wait_empty
        if stream_ack_cancelled_tasks is not None:
            warnings.warn(UserWarning(
                'deprecated stream_ack_cancelled_tasks will have no effect'))
            self.stream_ack_cancelled_tasks = stream_ack_cancelled_tasks
        if stream_ack_exceptions is not None:
            warnings.warn(UserWarning(
                'deprecated stream_ack_exceptions will have no effect'))
            self.stream_ack_exceptions = stream_ack_exceptions
        if stream_publish_on_commit is not None:
            self.stream_publish_on_commit = stream_publish_on_commit
        if stream_recovery_delay is not None:
            self.stream_recovery_delay = cast(float, stream_recovery_delay)
        if stream_processing_timeout is not None:
            self.stream_processing_timeout = cast(
                float, stream_processing_timeout)
        if producer_linger_ms is not None:
            self.producer_linger_ms = producer_linger_ms
        if producer_max_batch_size is not None:
            self.producer_max_batch_size = producer_max_batch_size
        if producer_acks is not None:
            self.producer_acks = producer_acks
        if producer_max_request_size is not None:
            self.producer_max_request_size = producer_max_request_size
        if producer_compression_type is not None:
            self.producer_compression_type = producer_compression_type
        if producer_partitioner is not None:
            self.producer_partitioner = producer_partitioner  # type: ignore
        if producer_request_timeout is not None:
            self.producer_request_timeout = cast(
                float, producer_request_timeout)
        if producer_api_version is not None:
            self.producer_api_version = producer_api_version
        if consumer_group_instance_id is not None:
            self.consumer_group_instance_id = consumer_group_instance_id
        if consumer_api_version is not None:
            self.consumer_api_version = consumer_api_version
        if consumer_max_fetch_size is not None:
            self.consumer_max_fetch_size = consumer_max_fetch_size
        if consumer_auto_offset_reset is not None:
            self.consumer_auto_offset_reset = consumer_auto_offset_reset
        if web_bind is not None:
            self.web_bind = web_bind
        if web_port is not None:
            self.web_port = web_port
        if web_host is not None:
            self.web_host = web_host
        if web_transport is not None:
            # XXX have to cast to silence mypy bug
            self.web_transport = cast(URL, web_transport)
        if web_in_thread is not None:
            self.web_in_thread = web_in_thread
        if web_cors_options is not None:
            self.web_cors_options = web_cors_options
        if worker_redirect_stdouts is not None:
            self.worker_redirect_stdouts = worker_redirect_stdouts
        if worker_redirect_stdouts_level is not None:
            self.worker_redirect_stdouts_level = worker_redirect_stdouts_level
        if worker_shutdown_memory_utilization_percent is not None:
            self.worker_shutdown_memory_utilization_percent = worker_shutdown_memory_utilization_percent
        if worker_shutdown_memory_utilization_bytes is not None:
            self.worker_shutdown_memory_utilization_bytes = worker_shutdown_memory_utilization_bytes

        if reply_to_prefix is not None:
            self.reply_to_prefix = reply_to_prefix
        if reply_to is not None:
            self.reply_to = reply_to
        else:
            self.reply_to = f'{self.reply_to_prefix}{uuid4()}'
        if reply_expires is not None:
            self.reply_expires = cast(float, reply_expires)

        self.GlobalTable = cast(
            Type[GlobalTableT], GlobalTable or GLOBAL_TABLE_TYPE)
        self.SetGlobalTable = cast(
            Type[GlobalTableT], SetGlobalTable or SET_GLOBAL_TABLE_TYPE)
        self.agent_supervisor = cast(
            Type[SupervisorStrategyT],
            agent_supervisor or AGENT_SUPERVISOR_TYPE)

        self.Agent = cast(Type[AgentT], Agent or AGENT_TYPE)
        self.ConsumerScheduler = cast(
            Type[SchedulingStrategyT],
            ConsumerScheduler or CONSUMER_SCHEDULER_TYPE)
        self.Event = cast(Type[EventT], Event or EVENT_TYPE)
        self.Schema = cast(Type[SchemaT], Schema or SCHEMA_TYPE)
        self.Stream = cast(Type[StreamT], Stream or STREAM_TYPE)
        self.Table = cast(Type[TableT], Table or TABLE_TYPE)
        self.SetTable = cast(Type[TableT], SetTable or SET_TABLE_TYPE)
        self.TableManager = cast(
            Type[TableManagerT],
            TableManager or TABLE_MANAGER_TYPE)
        self.Serializers = cast(Type[RegistryT], Serializers or REGISTRY_TYPE)
        self.Worker = cast(Type[_WorkerT], Worker or WORKER_TYPE)
        self.PartitionAssignor = cast(
            Type[PartitionAssignorT],
            PartitionAssignor or PARTITION_ASSIGNOR_TYPE)
        self.LeaderAssignor = cast(
            Type[LeaderAssignorT],
            LeaderAssignor or LEADER_ASSIGNOR_TYPE)
        self.Router = cast(Type[RouterT], Router or ROUTER_TYPE)
        self.Topic = cast(Type[TopicT], Topic or TOPIC_TYPE)
        self.HttpClient = cast(
            Type[HttpClientT], HttpClient or HTTP_CLIENT_TYPE)
        self.Monitor = cast(Type[SensorT], Monitor or MONITOR_TYPE)
        self.__dict__.update(kwargs)  # arbitrary configuration
        object.__setattr__(self, '_accessed', set())
        object.__setattr__(self, '_initializing', False)

    def __getattribute__(self, key: str) -> Any:
        accessed = object.__getattribute__(self, '_accessed')
        if not key.startswith('_'):
            if not object.__getattribute__(self, '_initializing'):
                accessed.add(key)
        return object.__getattribute__(self, key)

    def __setattr__(self, key: str, value: Any) -> None:
        xsd = object.__getattribute__(self, '_accessed')
        if xsd is not None and not key.startswith('_') and key in xsd:
            old_value = object.__getattribute__(self, key)
            self._warn_already_configured_key(key, value, old_value)
        object.__setattr__(self, key, value)

    def _first_not_none(self, *args: Any) -> Any:
        for v in args:
            if v is not None:
                return v

    def _prepare_id(self, id: str) -> str:
        if self.version > 1:
            return self.id_format.format(id=id, self=self)
        return id

    def _prepare_datadir(self, datadir: Union[str, Path]) -> Path:
        return self._Path(str(datadir).format(conf=self))

    def _prepare_tabledir(self, tabledir: Union[str, Path]) -> Path:
        return self._appdir_path(self._Path(tabledir))

    def _Path(self, *parts: Union[str, Path]) -> Path:
        return Path(*parts).expanduser()

    def _appdir_path(self, path: Path) -> Path:
        return path if path.is_absolute() else self.appdir / path

    @property
    def name(self) -> str:
        # name is a read-only property
        return self._name

    @property
    def id(self) -> str:
        return self._id

    @id.setter
    def id(self, name: str) -> None:
        self._name = name
        self._id = self._prepare_id(name)  # id is name+version

    @property
    def origin(self) -> Optional[str]:
        return self._origin

    @origin.setter
    def origin(self, origin: Optional[str]) -> None:
        self._origin = origin

    @property
    def version(self) -> int:
        return self._version

    @version.setter
    def version(self, version: int) -> None:
        if not version:
            raise ImproperlyConfigured(
                f'Version cannot be {version}, please start at 1')
        self._version = version

    @property
    def broker(self) -> List[URL]:
        return self._broker

    @broker.setter
    def broker(self, broker: Union[URL, str, List[URL]]) -> None:
        self._broker = urllist(broker, default_scheme=DEFAULT_BROKER_SCHEME)

    @property
    def broker_consumer(self) -> List[URL]:
        consumer = self._broker_consumer
        return consumer if consumer is not None else self.broker

    @broker_consumer.setter
    def broker_consumer(self, broker: Union[URL, str, List[URL]]) -> None:
        self._broker_consumer = urllist(
            broker, default_scheme=DEFAULT_BROKER_SCHEME)

    @property
    def broker_producer(self) -> List[URL]:
        producer = self._broker_producer
        return producer if producer is not None else self.broker

    @broker_producer.setter
    def broker_producer(self, broker: Union[URL, str, List[URL]]) -> None:
        self._broker_producer = urllist(
            broker, default_scheme=DEFAULT_BROKER_SCHEME)

    @property
    def store(self) -> URL:
        return self._store

    @store.setter
    def store(self, store: Union[URL, str]) -> None:
        self._store = URL(store)

    @property
    def web(self) -> URL:
        return self._web

    @web.setter
    def web(self, web: Union[URL, str]) -> None:
        self._web = URL(web)

    @property
    def cache(self) -> URL:
        return self._cache

    @cache.setter
    def cache(self, cache: Union[URL, str]) -> None:
        self._cache = URL(cache)

    @property
    def canonical_url(self) -> URL:
        if not self._canonical_url:
            self._canonical_url = URL(
                f'http://{self.web_host}:{self.web_port}')
        return self._canonical_url

    @canonical_url.setter
    def canonical_url(self, canonical_url: Union[URL, str]) -> None:
        self._canonical_url = URL(canonical_url)

    @property
    def datadir(self) -> Path:
        return self._datadir

    @datadir.setter
    def datadir(self, datadir: Union[Path, str]) -> None:
        self._datadir = self._prepare_datadir(datadir)

    @property
    def appdir(self) -> Path:
        return self._versiondir(self.version)

    def _versiondir(self, version: int) -> Path:
        return self.datadir / f'v{version}'

    def find_old_versiondirs(self) -> Iterable[Path]:
        for version in reversed(range(0, self.version)):
            path = self._versiondir(version)
            if path.is_dir():
                yield path

    @property
    def tabledir(self) -> Path:
        return self._tabledir

    @tabledir.setter
    def tabledir(self, tabledir: Union[Path, str]) -> None:
        self._tabledir = self._prepare_tabledir(tabledir)

    @property
    def processing_guarantee(self) -> ProcessingGuarantee:
        return self._processing_guarantee

    @processing_guarantee.setter
    def processing_guarantee(self,
                             value: Union[str, ProcessingGuarantee]) -> None:
        self._processing_guarantee = ProcessingGuarantee(value)

    @property
    def broker_credentials(self) -> Optional[CredentialsT]:
        return self._broker_credentials

    @broker_credentials.setter
    def broker_credentials(self, creds: CredentialsArg = None) -> None:
        self._broker_credentials = to_credentials(creds)

    @property
    def broker_request_timeout(self) -> float:
        return self._broker_request_timeout

    @broker_request_timeout.setter
    def broker_request_timeout(self, value: Seconds) -> None:
        self._broker_request_timeout = want_seconds(value)

    @property
    def broker_session_timeout(self) -> float:
        return self._broker_session_timeout

    @broker_session_timeout.setter
    def broker_session_timeout(self, value: Seconds) -> None:
        self._broker_session_timeout = want_seconds(value)

    @property
    def broker_rebalance_timeout(self) -> float:
        return self._broker_rebalance_timeout

    @broker_rebalance_timeout.setter
    def broker_rebalance_timeout(self, value: Seconds) -> None:
        self._broker_rebalance_timeout = want_seconds(value)

    @property
    def broker_heartbeat_interval(self) -> float:
        return self._broker_heartbeat_interval

    @broker_heartbeat_interval.setter
    def broker_heartbeat_interval(self, value: Seconds) -> None:
        self._broker_heartbeat_interval = want_seconds(value)

    @property
    def broker_commit_interval(self) -> float:
        return self._broker_commit_interval

    @broker_commit_interval.setter
    def broker_commit_interval(self, value: Seconds) -> None:
        self._broker_commit_interval = want_seconds(value)

    @property
    def broker_commit_livelock_soft_timeout(self) -> float:
        return self._broker_commit_livelock_soft_timeout

    @broker_commit_livelock_soft_timeout.setter
    def broker_commit_livelock_soft_timeout(self, value: Seconds) -> None:
        self._broker_commit_livelock_soft_timeout = want_seconds(value)

    @property
    def broker_max_poll_records(self) -> Optional[int]:
        return self._broker_max_poll_records

    @broker_max_poll_records.setter
    def broker_max_poll_records(self, value: Optional[int]) -> None:
        self._broker_max_poll_records = value

    @property
    def consumer_api_version(self) -> str:
        consumer_api_version = self._consumer_api_version
        if consumer_api_version is None:
            return self.broker_api_version
        else:
            return consumer_api_version

    @consumer_api_version.setter
    def consumer_api_version(self, version: str) -> None:
        self._consumer_api_version = version

    @property
    def producer_api_version(self) -> str:
        producer_api_version = self._producer_api_version
        if producer_api_version is None:
            return self.broker_api_version
        else:
            return producer_api_version

    @producer_api_version.setter
    def producer_api_version(self, version: str) -> None:
        self._producer_api_version = version

    @property
    def producer_partitioner(self) -> Optional[PartitionerT]:
        return self._producer_partitioner

    @producer_partitioner.setter
    def producer_partitioner(
            self, handler: Optional[SymbolArg[PartitionerT]]) -> None:
        self._producer_partitioner = symbol_by_name(handler)

    @property
    def producer_request_timeout(self) -> float:
        return self._producer_request_timeout

    @producer_request_timeout.setter
    def producer_request_timeout(self, timeout: Seconds) -> None:
        self._producer_request_timeout = want_seconds(timeout)

    @property
    def table_cleanup_interval(self) -> float:
        return self._table_cleanup_interval

    @table_cleanup_interval.setter
    def table_cleanup_interval(self, value: Seconds) -> None:
        self._table_cleanup_interval = want_seconds(value)

    @property
    def reply_expires(self) -> float:
        return self._reply_expires

    @reply_expires.setter
    def reply_expires(self, reply_expires: Seconds) -> None:
        self._reply_expires = want_seconds(reply_expires)

    @property
    def stream_recovery_delay(self) -> float:
        return self._stream_recovery_delay

    @stream_recovery_delay.setter
    def stream_recovery_delay(self, delay: Seconds) -> None:
        self._stream_recovery_delay = want_seconds(delay)

    @property
    def stream_processing_timeout(self) -> float:
        return self._stream_processing_timeout

    @stream_processing_timeout.setter
    def stream_processing_timeout(self, timeout: Seconds) -> None:
        self._stream_processing_timeout = want_seconds(timeout)

    @property
    def agent_supervisor(self) -> Type[SupervisorStrategyT]:
        return self._agent_supervisor

    @agent_supervisor.setter
    def agent_supervisor(
            self, sup: SymbolArg[Type[SupervisorStrategyT]]) -> None:
        self._agent_supervisor = symbol_by_name(sup)

    @property
    def web_transport(self) -> URL:
        return self._web_transport

    @web_transport.setter
    def web_transport(self, url: Union[str, URL]) -> None:
        self._web_transport = URL(url)

    @property
    def Agent(self) -> Type[AgentT]:
        return self._Agent

    @Agent.setter
    def Agent(self, Agent: SymbolArg[Type[AgentT]]) -> None:
        self._Agent = symbol_by_name(Agent)

    @property
    def ConsumerScheduler(self) -> Type[SchedulingStrategyT]:
        return self._ConsumerScheduler

    @ConsumerScheduler.setter
    def ConsumerScheduler(
            self, value: SymbolArg[Type[SchedulingStrategyT]]) -> None:
        self._ConsumerScheduler = symbol_by_name(value)

    @property
    def Event(self) -> Type[EventT]:
        return self._Event

    @Event.setter
    def Event(self, Event: SymbolArg[Type[EventT]]) -> None:
        self._Event = symbol_by_name(Event)

    @property
    def Schema(self) -> Type[SchemaT]:
        return self._Schema

    @Schema.setter
    def Schema(self, Schema: SymbolArg[Type[SchemaT]]) -> None:
        self._Schema = symbol_by_name(Schema)

    @property
    def Stream(self) -> Type[StreamT]:
        return self._Stream

    @Stream.setter
    def Stream(self, Stream: SymbolArg[Type[StreamT]]) -> None:
        self._Stream = symbol_by_name(Stream)

    @property
    def Table(self) -> Type[TableT]:
        return self._Table

    @Table.setter
    def Table(self, Table: SymbolArg[Type[TableT]]) -> None:
        self._Table = symbol_by_name(Table)

    @property
    def SetTable(self) -> Type[TableT]:
        return self._SetTable

    @SetTable.setter
    def SetTable(self, SetTable: SymbolArg[Type[TableT]]) -> None:
        self._SetTable = symbol_by_name(SetTable)

    @property
    def GlobalTable(self) -> Type[GlobalTableT]:
        return self._GlobalTable

    @GlobalTable.setter
    def GlobalTable(self, GlobalTable: SymbolArg[Type[GlobalTableT]]) -> None:
        self._GlobalTable = symbol_by_name(GlobalTable)

    @property
    def SetGlobalTable(self) -> Type[GlobalTableT]:
        return self._SetGlobalTable

    @SetGlobalTable.setter
    def SetGlobalTable(
            self, SetGlobalTable: SymbolArg[Type[GlobalTableT]]) -> None:
        self._SetGlobalTable = symbol_by_name(SetGlobalTable)

    @property
    def TableManager(self) -> Type[TableManagerT]:
        return self._TableManager

    @TableManager.setter
    def TableManager(self, Manager: SymbolArg[Type[TableManagerT]]) -> None:
        self._TableManager = symbol_by_name(Manager)

    @property
    def Serializers(self) -> Type[RegistryT]:
        return self._Serializers

    @Serializers.setter
    def Serializers(self, Serializers: SymbolArg[Type[RegistryT]]) -> None:
        self._Serializers = symbol_by_name(Serializers)

    @property
    def Worker(self) -> Type[_WorkerT]:
        return self._Worker

    @Worker.setter
    def Worker(self, Worker: SymbolArg[Type[_WorkerT]]) -> None:
        self._Worker = symbol_by_name(Worker)

    @property
    def PartitionAssignor(self) -> Type[PartitionAssignorT]:
        return self._PartitionAssignor

    @PartitionAssignor.setter
    def PartitionAssignor(
            self, Assignor: SymbolArg[Type[PartitionAssignorT]]) -> None:
        self._PartitionAssignor = symbol_by_name(Assignor)

    @property
    def LeaderAssignor(self) -> Type[LeaderAssignorT]:
        return self._LeaderAssignor

    @LeaderAssignor.setter
    def LeaderAssignor(
            self, Assignor: SymbolArg[Type[LeaderAssignorT]]) -> None:
        self._LeaderAssignor = symbol_by_name(Assignor)

    @property
    def Router(self) -> Type[RouterT]:
        return self._Router

    @Router.setter
    def Router(self, Router: SymbolArg[Type[RouterT]]) -> None:
        self._Router = symbol_by_name(Router)

    @property
    def Topic(self) -> Type[TopicT]:
        return self._Topic

    @Topic.setter
    def Topic(self, Topic: SymbolArg[Type[TopicT]]) -> None:
        self._Topic = symbol_by_name(Topic)

    @property
    def HttpClient(self) -> Type[HttpClientT]:
        return self._HttpClient

    @HttpClient.setter
    def HttpClient(self, HttpClient: SymbolArg[Type[HttpClientT]]) -> None:
        self._HttpClient = symbol_by_name(HttpClient)

    @property
    def Monitor(self) -> Type[SensorT]:
        return self._Monitor

    @Monitor.setter
    def Monitor(self, Monitor: SymbolArg[Type[SensorT]]) -> None:
        self._Monitor = symbol_by_name(Monitor)



from pathlib import Path
from typing import List, Mapping, Tuple
import pytest
from faust.exceptions import ImproperlyConfigured
from faust.stores import rocksdb
from faust.stores.rocksdb import RocksDBOptions, Store
from faust.types import TP
from mode.utils.mocks import AsyncMock, Mock, call, patch
from yarl import URL

TP1 = TP('foo', 0)
TP2 = TP('foo', 1)
TP3 = TP('bar', 2)
TP4 = TP('baz', 3)


class MockIterator(Mock):

    @classmethod
    def from_values(cls, values):
        it = cls()
        it.values = values
        return it

    def __iter__(self):
        return iter(self.values)


class test_RocksDBOptions:

    @pytest.mark.parametrize('arg', [
        'max_open_files',
        'write_buffer_size',
        'max_write_buffer_number',
        'target_file_size_base',
        'block_cache_size',
        'block_cache_compressed_size',
        'bloom_filter_size',
    ])
    def test_init(self, arg):
        opts = RocksDBOptions(**{arg: 30})
        assert getattr(opts, arg) == 30

    def test_defaults(self):
        opts = RocksDBOptions()
        assert opts.max_open_files == rocksdb.DEFAULT_MAX_OPEN_FILES
        assert opts.write_buffer_size == rocksdb.DEFAULT_WRITE_BUFFER_SIZE
        assert (opts.max_write_buffer_number ==
                rocksdb.DEFAULT_MAX_WRITE_BUFFER_NUMBER)
        assert (opts.target_file_size_base ==
                rocksdb.DEFAULT_TARGET_FILE_SIZE_BASE)
        assert opts.block_cache_size == rocksdb.DEFAULT_BLOCK_CACHE_SIZE
        assert (opts.block_cache_compressed_size ==
                rocksdb.DEFAULT_BLOCK_CACHE_COMPRESSED_SIZE)
        assert opts.bloom_filter_size == rocksdb.DEFAULT_BLOOM_FILTER_SIZE

    def test_open(self):
        with patch('faust.stores.rocksdb.rocksdb', Mock()) as rocks:
            opts = RocksDBOptions()
            db = opts.open(Path('foo.db'), read_only=True)
            rocks.DB.assert_called_once_with(
                'foo.db', opts.as_options(), read_only=True)
            assert db is rocks.DB()


class test_Store:

    @pytest.fixture()
    def table(self):
        table = Mock(name='table')
        table.name = 'table1'
        return table

    @pytest.yield_fixture()
    def rocks(self):
        with patch('faust.stores.rocksdb.rocksdb') as rocks:
            yield rocks

    @pytest.yield_fixture()
    def no_rocks(self):
        with patch('faust.stores.rocksdb.rocksdb', None) as rocks:
            yield rocks

    @pytest.fixture()
    def store(self, *, app, rocks, table):
        return Store('rocksdb://', app, table)

    @pytest.fixture()
    def db_for_partition(self, *, store):
        dfp = store._db_for_partition = Mock(name='db_for_partition')
        return dfp

    def test_default_key_index_size(self, *, store):
        assert store.key_index_size == store.app.conf.table_key_index_size

    def test_set_key_index_size(self, *, app, rocks, table):
        s = Store('rocksdb://', app, table, key_index_size=12341)

        assert s.key_index_size == 12341

    def test_no_rocksdb(self, *, app, table, no_rocks):
        with pytest.raises(ImproperlyConfigured):
            Store('rocksdb://', app, table)

    def test_url_without_path_adds_table_name(self, *, store):
        assert store.url == URL('rocksdb:table1')

    def test_url_having_path(self, *, app, rocks, table):
        store = Store('rocksdb://foobar/', app, table)
        assert store.url == URL('rocksdb://foobar/')

    def test_init(self, *, store, app):
        assert isinstance(store.rocksdb_options, RocksDBOptions)
        assert store.key_index_size == app.conf.table_key_index_size
        assert store._dbs == {}
        assert store._key_index is not None

    def test_persisted_offset(self, *, store, db_for_partition):
        db_for_partition.return_value.get.return_value = '300'
        assert store.persisted_offset(TP1) == 300
        db_for_partition.assert_called_once_with(TP1.partition)
        db_for_partition.return_value.get.assert_called_once_with(
            store.offset_key)

        db_for_partition.return_value.get.return_value = None
        assert store.persisted_offset(TP1) is None

    def test_set_persisted_offset(self, *, store, db_for_partition):
        store.set_persisted_offset(TP1, 3003)
        db_for_partition.assert_called_once_with(TP1.partition)
        db_for_partition.return_value.put.assert_called_once_with(
            store.offset_key, b'3003',
        )

    @pytest.mark.asyncio
    async def test_need_active_standby_for(self, *, store, db_for_partition):
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):
            db_for_partition.side_effect = KeyError('lock acquired')
            assert not await store.need_active_standby_for(TP1)

    @pytest.mark.asyncio
    async def test_need_active_standby_for__raises(
            self, *, store, db_for_partition):
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):
            db_for_partition.side_effect = KeyError('oh no')
            with pytest.raises(KeyError):
                await store.need_active_standby_for(TP1)

    @pytest.mark.asyncio
    async def test_need_active_standby_for__active(
            self, *, store, db_for_partition):
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):
            assert await store.need_active_standby_for(TP1)

    def test_apply_changelog_batch(self, *, store, rocks, db_for_partition):

        def new_event(name, tp: TP, offset, key, value) -> Mock:
            return Mock(
                name='event1',
                message=Mock(
                    tp=tp,
                    topic=tp.topic,
                    partition=tp.partition,
                    offset=offset,
                    key=key,
                    value=value,
                ),
            )

        events = [
            new_event('event1', TP1, 1001, 'k1', 'v1'),
            new_event('event2', TP2, 2002, 'k2', 'v2'),
            new_event('event3', TP3, 3003, 'k3', 'v3'),
            new_event('event4', TP4, 4004, 'k4', 'v4'),
            new_event('event5', TP4, 4005, 'k5', None),
        ]

        dbs = {
            TP1.partition: Mock(name='db1'),
            TP2.partition: Mock(name='db2'),
            TP3.partition: Mock(name='db3'),
            TP4.partition: Mock(name='db4'),
        }
        db_for_partition.side_effect = dbs.get

        store.set_persisted_offset = Mock(name='set_persisted_offset')

        store.apply_changelog_batch(events, None, None)

        rocks.WriteBatch.return_value.delete.assert_called_once_with('k5')
        rocks.WriteBatch.return_value.put.assert_has_calls([
            call('k1', 'v1'),
            call('k2', 'v2'),
            call('k3', 'v3'),
            call('k4', 'v4'),
        ])

        for db in dbs.values():
            db.write.assert_called_once_with(rocks.WriteBatch())

        store.set_persisted_offset.assert_has_calls([
            call(TP1, 1001),
            call(TP2, 2002),
            call(TP3, 3003),
            call(TP4, 4005),
        ])

    @pytest.yield_fixture()
    def current_event(self):
        with patch('faust.stores.rocksdb.current_event') as current_event:
            yield current_event.return_value

    def test__set(self, *, store, db_for_partition, current_event):
        store._set(b'key', b'value')
        db_for_partition.assert_called_once_with(
            current_event.message.partition)
        assert store._key_index[b'key'] == current_event.message.partition
        db_for_partition.return_value.put.assert_called_once_with(
            b'key', b'value',
        )

    def test_db_for_partition(self, *, store):
        ofp = store._open_for_partition = Mock(name='open_for_partition')

        assert store._db_for_partition(1) is ofp.return_value
        assert store._dbs[1] is ofp.return_value

        assert store._db_for_partition(1) is ofp.return_value

        ofp.assert_called_once_with(1)

    def test_open_for_partition(self, *, store):
        open = store.rocksdb_options.open = Mock(name='options.open')
        assert store._open_for_partition(1) is open.return_value
        open.assert_called_once_with(store.partition_path(1))

    def test__get__missing(self, *, store):
        store._get_bucket_for_key = Mock(name='get_bucket_for_key')
        store._get_bucket_for_key.return_value = None
        assert store._get(b'key') is None

    def test__get(self, *, store):
        db = Mock(name='db')
        value = b'foo'
        store._get_bucket_for_key = Mock(name='get_bucket_for_key')
        store._get_bucket_for_key.return_value = (db, value)

        assert store._get(b'key') == value

    def test__get__dbvalue_is_None(self, *, store):
        db = Mock(name='db')
        store._get_bucket_for_key = Mock(name='get_bucket_for_key')
        store._get_bucket_for_key.return_value = (db, None)

        db.key_may_exist.return_value = [False]
        assert store._get(b'key') is None

        db.key_may_exist.return_value = [True]
        db.get.return_value = None
        assert store._get(b'key') is None

        db.get.return_value = b'bar'
        assert store._get(b'key') == b'bar'

    def test_get_bucket_for_key__is_in_index(self, *, store):
        store._key_index[b'key'] = 30
        db = store._dbs[30] = Mock(name='db-p30')

        db.key_may_exist.return_value = [False]
        assert store._get_bucket_for_key(b'key') is None

        db.key_may_exist.return_value = [True]
        db.get.return_value = None
        assert store._get_bucket_for_key(b'key') is None

        db.get.return_value = b'value'
        assert store._get_bucket_for_key(b'key') == (db, b'value')

    def test_get_bucket_for_key__no_dbs(self, *, store):
        assert store._get_bucket_for_key(b'key') is None

    def new_db(self, name, exists=False):
        db = Mock(name=name)
        db.key_may_exist.return_value = [exists]
        db.get.return_value = name
        return db

    def test_get_bucket_for_key__not_in_index(self, *, store):

        dbs = {
            1: self.new_db(name='db1'),
            2: self.new_db(name='db2'),
            3: self.new_db(name='db3', exists=True),
            4: self.new_db(name='db4', exists=True),
        }
        store._dbs.update(dbs)

        assert store._get_bucket_for_key(b'key') == (dbs[3], 'db3')

    def test__del(self, *, store):
        dbs = store._dbs_for_key = Mock(return_value=[
            Mock(name='db1'),
            Mock(name='db2'),
            Mock(name='db3'),
        ])
        store._del(b'key')
        for db in dbs.return_value:
            db.delete.assert_called_once_with(b'key')

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, store, table):
        store.revoke_partitions = Mock(name='revoke_partitions')
        store.assign_partitions = AsyncMock(name='assign_partitions')

        assigned = {TP1, TP2}
        revoked = {TP3}
        newly_assigned = {TP2}
        await store.on_rebalance(table, assigned, revoked, newly_assigned)

        store.revoke_partitions.assert_called_once_with(table, revoked)
        store.assign_partitions.assert_called_once_with(table, newly_assigned)

    def test_revoke_partitions(self, *, store, table):
        table.changelog_topic.topics = {TP1.topic, TP3.topic}
        store._dbs[TP3.partition] = Mock(name='db')

        with patch('gc.collect') as collect:
            store.revoke_partitions(table, {TP1, TP2, TP3, TP4})
            assert not store._dbs

            collect.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_assign_partitions(self, *, store, app, table):
        app.assignor.assigned_standbys = Mock(return_value={TP4})
        table.changelog_topic.topics = list({
            tp.topic for tp in (TP1, TP2, TP4)
        })

        store._try_open_db_for_partition = AsyncMock()
        await store.assign_partitions(table, {TP1, TP2, TP3, TP4})
        store._try_open_db_for_partition.assert_has_calls([
            call(TP2.partition),
            call.coro(TP2.partition),
            call(TP1.partition),
            call.coro(TP1.partition),
        ], any_order=True)

    @pytest.mark.asyncio
    async def test_assign_partitions__empty_assignment(
            self, *, store, app, table):
        app.assignor.assigned_standbys = Mock(return_value={TP4})
        table.changelog_topic.topics = list({
            tp.topic for tp in (TP1, TP2, TP4)
        })
        await store.assign_partitions(table, set())

    @pytest.mark.asyncio
    async def test_open_db_for_partition(self, *, store, db_for_partition):
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):
            assert (await store._try_open_db_for_partition(3) is
                    db_for_partition.return_value)

    @pytest.mark.asyncio
    async def test_open_db_for_partition_max_retries(
            self, *, store, db_for_partition):
        store.sleep = AsyncMock(name='sleep')
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):
            db_for_partition.side_effect = KeyError('lock already')
            with pytest.raises(KeyError):
                await store._try_open_db_for_partition(3)
        assert store.sleep.call_count == 4

    @pytest.mark.asyncio
    async def test_open_db_for_partition__raises_unexpected_error(
            self, *, store, db_for_partition):
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):
            db_for_partition.side_effect = RuntimeError()
            with pytest.raises(RuntimeError):
                await store._try_open_db_for_partition(3)

    @pytest.mark.asyncio
    async def test_open_db_for_partition_retries_recovers(
            self, *, store, db_for_partition):
        with patch('faust.stores.rocksdb.rocksdb.errors.RocksIOError',
                   KeyError):

            def on_call(partition):
                if db_for_partition.call_count < 3:
                    raise KeyError('lock already')
            db_for_partition.side_effect = on_call

            await store._try_open_db_for_partition(3)

    def test__contains(self, *, store):
        db1 = self.new_db('db1', exists=False)
        db2 = self.new_db('db2', exists=True)
        dbs = {b'key': [db1, db2]}
        store._dbs_for_key = Mock(side_effect=dbs.get)

        db2.get.return_value = None
        assert not store._contains(b'key')

        db2.get.return_value = b'value'
        assert store._contains(b'key')

    def test__dbs_for_key(self, *, store):
        dbs = store._dbs = {
            1: self.new_db('db1'),
            2: self.new_db('db2'),
            3: self.new_db('db3'),
        }
        store._key_index[b'key'] = 2

        assert list(store._dbs_for_key(b'other')) == list(dbs.values())
        assert list(store._dbs_for_key(b'key')) == [dbs[2]]

    def test__dbs_for_actives(self, *, store, table):
        table._changelog_topic_name.return_value = 'clog'
        store.app.assignor.assigned_actives = Mock(return_value=[
            TP('clog', 1),
            TP('clog', 2),
        ])
        dbs = store._dbs = {
            1: self.new_db('db1'),
            2: self.new_db('db2'),
            3: self.new_db('db3'),
        }

        # Normal Table
        table.is_global = False
        assert list(store._dbs_for_actives()) == [dbs[1], dbs[2]]

        # Global Table
        table.is_global = True
        assert list(store._dbs_for_actives()) == [dbs[1], dbs[2], dbs[3]]

    def test__size(self, *, store):
        dbs = self._setup_keys(
            db1=[
                store.offset_key,
                b'foo',
                b'bar',
            ],
            db2=[
                b'baz',
                store.offset_key,
                b'xuz',
                b'xaz',
            ],
        )
        store._dbs_for_actives = Mock(return_value=dbs)
        assert store._size() == 5

    def test__iterkeys(self, *, store):
        dbs = self._setup_keys(
            db1=[
                store.offset_key,
                b'foo',
                b'bar',
            ],
            db2=[
                b'baz',
                store.offset_key,
                b'xuz',
            ],
        )
        store._dbs_for_actives = Mock(return_value=dbs)

        assert list(store._iterkeys()) == [
            b'foo',
            b'bar',
            b'baz',
            b'xuz',
        ]

        for db in dbs:
            db.iterkeys.assert_called_once_with()
            db.iterkeys().seek_to_first.assert_called_once_with()

    def _setup_keys(self, **dbs: Mapping[str, List[bytes]]):
        return [
            self._setup_keys_db(name, values)
            for name, values in dbs.items()
        ]

    def _setup_keys_db(self, name: str, values: List[bytes]):
        db = self.new_db(name)
        db.iterkeys.return_value = MockIterator.from_values(values)
        return db

    def test__itervalues(self, *, store):
        dbs = self._setup_items(
            db1=[
                (store.offset_key, b'1001'),
                (b'k1', b'foo'),
                (b'k2', b'bar'),
            ],
            db2=[
                (b'k3', b'baz'),
                (store.offset_key, b'2002'),
                (b'k4', b'xuz'),
            ],
        )
        store._dbs_for_actives = Mock(return_value=dbs)

        assert list(store._itervalues()) == [
            b'foo',
            b'bar',
            b'baz',
            b'xuz',
        ]

        for db in dbs:
            db.iteritems.assert_called_once_with()
            db.iteritems().seek_to_first.assert_called_once_with()

    def _setup_items(self, **dbs: Mapping[str, List[Tuple[bytes, bytes]]]):
        return [
            self._setup_items_db(name, values)
            for name, values in dbs.items()
        ]

    def _setup_items_db(self, name: str, values: List[Tuple[bytes, bytes]]):
        db = self.new_db(name)
        db.iteritems.return_value = MockIterator.from_values(values)
        return db

    def test__iteritems(self, *, store):
        dbs = self._setup_items(
            db1=[
                (store.offset_key, b'1001'),
                (b'k1', b'foo'),
                (b'k2', b'bar'),
            ],
            db2=[
                (b'k3', b'baz'),
                (store.offset_key, b'2002'),
                (b'k4', b'xuz'),
            ],
        )
        store._dbs_for_actives = Mock(return_value=dbs)

        assert list(store._iteritems()) == [
            (b'k1', b'foo'),
            (b'k2', b'bar'),
            (b'k3', b'baz'),
            (b'k4', b'xuz'),
        ]

        for db in dbs:
            db.iteritems.assert_called_once_with()
            db.iteritems().seek_to_first.assert_called_once_with()

    def test_clear(self, *, store):
        with pytest.raises(NotImplementedError):
            store._clear()

    def test_reset_state(self, *, store):
        with patch('shutil.rmtree') as rmtree:
            store.reset_state()
            rmtree.assert_called_once_with(store.path.absolute())



from faust.web import exceptions


def test_WebError():
    exc = exceptions.WebError(detail='detail', code=400)
    assert exc.detail == 'detail'
    assert exc.code == 400


def test_WebError_defaults():
    exc = exceptions.WebError()
    assert exc.detail == exceptions.WebError.detail
    assert exc.code is None



"""LiveCheck - Models."""
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional

from mode.utils.compat import want_str
from mode.utils.objects import cached_property
from mode.utils.text import abbr

from faust import Record
from faust.utils.iso8601 import parse as parse_iso8601

__all__ = ['State', 'SignalEvent', 'TestExecution', 'TestReport']

HEADER_TEST_ID = 'LiveCheck-Test-Id'
HEADER_TEST_NAME = 'LiveCheck-Test-Name'
HEADER_TEST_TIMESTAMP = 'LiveCheck-Test-Timestamp'
HEADER_TEST_EXPIRES = 'LiveCheck-Test-Expires'


class State(Enum):
    """Test execution status."""

    INIT = 'INIT'
    PASS = 'PASS'
    FAIL = 'FAIL'
    ERROR = 'ERROR'
    TIMEOUT = 'TIMEOUT'
    STALL = 'STALL'
    SKIP = 'SKIP'

    def is_ok(self) -> bool:
        """Return :const:`True` if this is considered an OK state."""
        return self in OK_STATES


OK_STATES = frozenset({State.INIT, State.PASS, State.SKIP})


class SignalEvent(Record):
    """Signal sent to test (see :class:`faust.livecheck.signals.Signal`)."""

    signal_name: str
    case_name: str
    key: Any
    value: Any


class TestExecution(Record, isodates=True):
    """Requested test execution."""

    id: str
    case_name: str
    timestamp: datetime
    test_args: List[Any]
    test_kwargs: Dict[str, Any]
    expires: datetime

    @classmethod
    def from_headers(cls, headers: Mapping) -> Optional['TestExecution']:
        """Create instance from mapping of HTTP/Kafka headers."""
        try:
            test_id = want_str(headers[HEADER_TEST_ID])
        except KeyError:
            return None
        else:
            test_name = headers[HEADER_TEST_NAME]
            timestamp = headers[HEADER_TEST_TIMESTAMP]
            expires = headers[HEADER_TEST_EXPIRES]

            return cls(
                id=test_id,
                case_name=want_str(test_name),
                timestamp=parse_iso8601(want_str(timestamp)),
                expires=parse_iso8601(want_str(expires)),
                test_args=(),
                test_kwargs={},
            )

    def as_headers(self) -> Mapping:
        """Return test metadata as mapping of HTTP/Kafka headers."""
        return {
            HEADER_TEST_ID: self.id,
            HEADER_TEST_NAME: self.case_name,
            HEADER_TEST_TIMESTAMP: self.timestamp.isoformat(),
            HEADER_TEST_EXPIRES: self.expires.isoformat(),
        }

    @cached_property
    def ident(self) -> str:
        """Return long identifier for this test used in logs."""
        return self._build_ident(self.case_name, self.id)

    @cached_property
    def shortident(self) -> str:
        """Return short identifier for this test used in logs."""
        return self._build_ident(
            self.short_case_name,
            abbr(self.id, max=15, suffix='[...]'),
        )

    def _build_ident(self, case_name: str, id: str) -> str:
        return f'{case_name}:{id}'

    def _now(self) -> datetime:
        return datetime.utcnow().astimezone(timezone.utc)

    @cached_property
    def human_date(self) -> str:
        """Return human-readable description of test timestamp."""
        if self.was_issued_today:
            return f'''Today {self.timestamp.strftime('%H:%M:%S')}'''
        else:
            return str(self.timestamp)

    @cached_property
    def was_issued_today(self) -> bool:
        """Return :const:`True` if test was issued on todays date."""
        return self.timestamp.date() == self._now().date()

    @cached_property
    def is_expired(self) -> bool:
        """Return :const:`True` if this test already expired."""
        return self._now() >= self.expires

    @cached_property
    def short_case_name(self) -> str:
        """Return abbreviated case name."""
        return self.case_name.split('.')[-1]


class TestReport(Record):
    """Report after test execution."""

    case_name: str
    state: State
    test: Optional[TestExecution]
    runtime: Optional[float]
    signal_latency: Dict[str, float]
    error: Optional[str]
    traceback: Optional[str]



"""examples/service.py

This examples starts a separate :pypi:`mode` service with the app.

If you want the service instance to be generally available
you may create a subclass of app, to define a new app.myservice attribute:

.. sourcecode:: python

    class App(faust.App):
        myservice: MyService

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            self.myservice = self.service(
                MyService(loop=self.loop, beacon=self.beacon),
            )

    app = App('service-example')
"""
import faust


app = faust.App('service-example')


@app.service
class MyService(faust.Service):

    async def on_start(self) -> None:
        self.log.info('STARTED')

    async def on_stop(self) -> None:
        self.log.info('STOPPED')


@app.agent(value_type=str)
async def consumer(stream):
    async for message in stream:
        print(f'Received: {message!r}')


@app.timer(1.0)
async def producer():
    await consumer.send('hello')


if __name__ == '__main__':
    app.main()
